class Atom:
    """
    An atom class consists of a predicate, a tuple of terms (constants or variables) and a temporal interval,
    e.g., A(X, mike)@[1,2] or A(X, mike)
    """
    def __init__(self, predicate, entity=None, interval=None):
        self.predicate = predicate
        self.entity = entity
        self.interval = interval

    def __eq__(self, other):
        if isinstance(other, Atom):
            if other.interval is None and self.interval is None:
                return other.predicate == self.predicate and other.entity == self.entity
            elif other.interval is not None and self.interval is not None:
                return other.predicate == self.predicate and other.entity == self.entity and str(other.interval) == str(self.interval)
            else:
                return False
        else:
            return False

    def get_predicate(self):
        return self.predicate

    def get_entity(self):
        return self.entity

    def set_entity(self, entity):
        self.entity = entity

    def __str__(self):
        if self.interval:
            if len(self.entity) > 1 or  self.entity[0].name != "nan":
                return self.predicate + "(" + ",".join([str(item) for item in self.entity]) + ")@" + str(self.interval)
            else:
                return self.predicate + "@" + str(self.interval)

        else:
            if self.entity:
                if len(self.entity) == 1 and self.entity[0].name == "nan":
                    return self.predicate

                return self.predicate + "(" + ",".join([str(item) for item in self.entity]) + ")"
            else:
                return self.predicate

    def __hash__(self):
        if self.entity:
            if len(self.entity) == 1 and self.entity[0].name == "nan":
                return hash("Atom" + str(self.predicate) + str(self.interval))

            return hash("Atom" + str(self.predicate) + ",".join([str(item) for item in self.entity])
                        + str(self.interval))
        else:
            return hash("Atom" + str(self.predicate) + str(self.interval))




from meteor_reasoner.materialization.index_build import build_index
from meteor_reasoner.utils.loader import *
from meteor_reasoner.utils.operate_dataset import get_min_max_rational
from meteor_reasoner.materialization.coalesce import coalescing_d, coalescing
from meteor_reasoner.materialization.t_operator import naive_immediate_consequence_operator
import time
from meteor_reasoner.materialization.materialize import naive_combine
import pickle
import sys
import decimal
import argparse
from meteor_reasoner.stream.utils import *
import logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)

output_file_handler = logging.FileHandler("./log/mock.log")
stdout_handler = logging.StreamHandler(sys.stdout)
logger.addHandler(output_file_handler)
logger.addHandler(stdout_handler)

parser = argparse.ArgumentParser()
parser.add_argument("--datapath", default="../datasets/jsw/traffic.txt",  type=str, help="Input the dataset path (json format)")
parser.add_argument("--rulepath", default="./programs/traffic_1", type=str, help="Input the program path")
parser.add_argument("--datasetname", default="local_lubm5_program_1", type=str, help="Input the save path")
parser.add_argument("--target", default="a1:Scientist", type=str, help="Input the program path")
args = parser.parse_args()

target_predicate = args.target
time_points = [decimal.Decimal(i) for i in range(1001)]
with open(args.rulepath) as file:
    raw_program = file.readlines()
    program = load_program(raw_program)
    predicates = set()
    for rule in program:
        predicates.add(rule.head.get_predicate())
        for literal in rule.body:
            if isinstance(literal, BinaryLiteral):
                predicates.add(literal.left_literal.get_predicate())
                predicates.add(literal.right_literal.get_predicate())
            else:
                predicates.add(literal.get_predicate())

start_time = time.time()
with open(args.datapath) as file:
    raw_data = []
    for line in file:
        for predicate in predicates:
            if line.startswith(predicate):
                 raw_data.append(line)
                 break
    print("Lines of facts:", len(raw_data))
    D = load_dataset(raw_data)
    for predicate in D:
        for entity in D[predicate]:
            interval_list = []
            for item in D[predicate][entity]:
                value = (item.left_value + item.right_value) / 2
                interval_list.append(Interval(value, value, False, False))
            D[predicate][entity] = interval_list

    coalescing_d(D)
    D_index = build_index(D)

print("Loading ends={}s".format(time.time()-start_time))
coalescing_d(D)
min_val, max_val = get_min_max_rational(D)
limit = Interval(min_val, max_val, False, False)
print("Min_value={}, Max_value={} in the given dataset".format(min_val, max_val))

start_time = time.time()
cnt = 0
window_sizes = []
window_sizes_raw = []
while True:
    cnt += 1
    print("cnt={}, {}s".format(cnt, time.time()-start_time))
    delta_new = naive_immediate_consequence_operator(D=D, rules=program, D_index=D_index)

    # drop facts out of the range
    delta_new_prime = trim_delta(D, delta_new, limit=limit)

    if len(delta_new_prime) != 0:
        num_facts_raw = 0

        # before coalescing
        for predicate in D:
            for entity in D[predicate]:
                num_facts_raw += len(D[predicate][entity])
        for predicate in delta_new_prime:
            for entity in delta_new_prime[predicate]:
                num_facts_raw += len(delta_new_prime[predicate][entity])

        window_sizes_raw.append(num_facts_raw)
        logger.info("Number of window facts: {}".format(num_facts_raw))
        # before coalescing

        fixpoint = naive_combine(D, delta_new_prime, D_index)

        num_facts = 0
        for predicate in D:
            for entity in D[predicate]:
                num_facts += len(D[predicate][entity])
        window_sizes.append(num_facts)

        logger.info("Number of window facts: uncoalesced={}, coalesced={}".format(num_facts_raw, num_facts))
        if fixpoint:
            break
    else:
        break

time_points = []
point = min_val - decimal.Decimal(0.5)
while point <= max_val:
    point += decimal.Decimal(0.5)
    time_points.append(point)

streams_out = 0
for point in time_points:
    for entity in D[target_predicate]:
        for interval in D[predicate][entity]:
            if interval.left_value < point < interval.right_value or (interval.left_value == point
                                                                      and not interval.left_open) or \
                    (interval.right_value == point and not interval.right_open):
                streams_out += 1
                break

run_times = time.time() - start_time
results = {
    "run_times": run_times,
    "window_size": window_sizes,
    "window_size_raw": window_sizes_raw,
    "stream_out": streams_out,
    "min_value": min_val,
    "max_value": max_val
}

logger.info(results)
from datetime import datetime
now = datetime.now()
output_path = "./results/meteor_" + args.datasetname + ".pkl"
pickle.dump(results, open(output_path, "wb"))
print("============Stream Reasoning Ends============\n")






from collections import deque


def generate_guess_points(w, binary_literals, ruler_intervals, z, gcd):
    to_guess_points = set()
    for interval in ruler_intervals:
        to_guess_points.add(interval.left_value)
        to_guess_points.add(interval.right_value)

    left_z = z
    right_z = z

    left_len_list = deque(
        [str(abs(interval.left_value - interval.right_value)) for interval in ruler_intervals if
         interval.left_value - interval.right_value != 0])
    right_len_list = deque(
        [str(abs(interval.left_value - interval.right_value)) for interval in ruler_intervals if
         interval.left_value - interval.right_value != 0])

    while left_z > 0:
        last_endpoint = ruler_intervals[0].left_value
        while True:
            pattern = "#".join(left_len_list)
            if pattern in w.left_pattern:
                next_len = w.left_pattern[pattern]
                if next_len <= left_z:
                    last_endpoint = last_endpoint - next_len
                    to_guess_points.add(last_endpoint)
                    left_len_list.appendleft(str(next_len))
                    left_z -= next_len
                    break

            del left_len_list[-1]

    while right_z > 0:
        last_endpoint = ruler_intervals[-1].right_value
        while True:
            pattern = "#".join(right_len_list)
            if pattern in w.left_pattern:
                next_len = w.left_pattern[pattern]
                if next_len <= right_z:
                    last_endpoint = last_endpoint + next_len
                    to_guess_points.add(last_endpoint)
                    right_len_list.appendleft(str(next_len))
                    right_z -= next_len
                    break

            del right_len_list[0]

    to_guess_points = sorted(list(to_guess_points))
    the_left_most_z = to_guess_points[0]
    the_right_most_z = to_guess_points[-1]

    left_right_literals = list()
    for item in binary_literals:
        if item.left_atom.get_predicate() not in ["Bottom", "Top"] and item.left_atom not in left_right_literals:
            left_right_literals.append(item.left_atom)
        if item.right_atom.get_predicate() not in ["Bottom", "Top"] and item.right_atom not in left_right_literals:
            left_right_literals.append(item.right_atom)

    for i in range(1, len(left_right_literals) + 1):
        to_guess_points.insert(0, the_left_most_z - gcd * i)
        to_guess_points.insert(-1, the_right_most_z + gcd * i)

    return to_guess_points, left_right_literals



from collections import defaultdict


def build_index(D, D_index=None):
    """
    Implement an index construction, assuming that the entity only contains at most two terms
    Args:
        D (dictionary of dictionary):
    Returns:
        Dictionary of Dictionary
    """
    if D_index is None:
        D_index = defaultdict(lambda: defaultdict(list))
        for predicate in D.keys():
            for entity, intervals in D[predicate].items():
                if len(entity) == 1:
                    break
                else:
                    for i, item in enumerate(entity):
                        D_index[predicate][str(i) + "@" + item.name].append(entity)

                    if len(entity) > 2:
                        for i, item1 in enumerate(entity):
                            for j, item2 in enumerate(entity):
                                if j <= i:
                                    continue
                                D_index[predicate][str(i) + "@" + item1.name+"||"+str(j) + "@" + item2.name].append(entity)
        return D_index
    else:
        for predicate in D.keys():
            for entity, intervals in D[predicate].items():
                if len(entity) == 1:
                    break
                else:
                    for i, item in enumerate(entity):
                        D_index[predicate][str(i) + "@" + item.name].append(entity)
                    if len(entity) > 2:
                        for i, item1 in enumerate(entity):
                            for j, item2 in enumerate(entity):
                                if j <= i:
                                    continue
                                D_index[predicate][str(i) + "@" + item1.name+"||"+str(j) + "@" + item2.name].append(entity)







from meteor_reasoner.stream.utils import *
import time


class Stream_Generator:
    def __init__(self, time_points, Data):
        self.time_points = time_points
        self.streams_data = defaultdict(list)
        for time_point in time_points:
            for predicate in Data:
                for entity, intervals in Data[predicate].items():
                    for interval in intervals:
                        if interval.left_value == time_point:
                            self.streams_data[time_point].append(Atom(predicate, entity))
                            break


    def generator(self, delay=0.001):
        for t in self.time_points:
            yield self.streams_data[t], t




from meteor_reasoner.classes.term import *
from meteor_reasoner.classes.rule import *
from collections import defaultdict
from meteor_reasoner.materialization.coalesce import coalescing
from meteor_reasoner.materialization.ground import *


def join(rule, D, delta_old, delta_new, common_fragment=None, D_index=None):
    """
    This function implement the join operator when variables exist in the body of the rule.
    Args:
        rule (a Rule instance):
        Delta (a dictionary object): store new materialized results.
        D (a dictionary object): a global variable storing all facts.

    Returns:
        None
    """
    head_entity = rule.head.get_entity()
    head_predicate = rule.head.get_predicate()
    literals = rule.body

    def ground_body(global_literal_index, visited, delta, context):
        if global_literal_index == len(literals):
            T = []
            for i in range(len(literals)):
                tmp_D = defaultdict(lambda: defaultdict(list))
                grounded_literal = copy.deepcopy(literals[i])
                if isinstance(grounded_literal, BinaryLiteral):
                    left_predicate = grounded_literal.left_atom.get_predicate()
                    right_predicate = grounded_literal.right_atom.get_predicate()
                    if left_predicate == "Top":
                        grounded_literal.right_atom.set_entity(delta[i][0])
                        tmp_D[grounded_literal.right_atom.get_predicate()][delta[i][0]] = delta[i][1]
                    elif right_predicate == "Top":
                        grounded_literal.left_atom.set_entity(delta[i][0])
                        tmp_D[grounded_literal.left_atom.get_predicate()][delta[i][0]] = delta[i][1]
                    else:
                        grounded_literal.left_atom.set_entity(delta[i][0])
                        tmp_D[grounded_literal.left_atom.get_predicate()][delta[i][0]] = delta[i][1]
                        grounded_literal.right_atom.set_entity(delta[i][2])
                        tmp_D[grounded_literal.right_atom.get_predicate()][delta[i][2]] = delta[i][3]

                else:
                    if grounded_literal.get_predicate() not in ["Bottom", "Top"]:
                        grounded_literal.set_entity(delta[i][0])
                        tmp_D[grounded_literal.get_predicate()][delta[i][0]] = delta[i][1]

                t = apply(grounded_literal, tmp_D)
                if len(t) == 0:
                    break
                else:
                    T.append(t)

            n_T = []
            if len(rule.negative_body_atoms) != 0:
                n_literals = rule.negative_body_atoms
                for i in range(len(n_literals)):
                    tmp_D = defaultdict(lambda: defaultdict(list))
                    grounded_literal = copy.deepcopy(n_literals[i])
                    if isinstance(grounded_literal, BinaryLiteral):
                        left_predicate = grounded_literal.left_atom.get_predicate()
                        right_predicate = grounded_literal.right_atom.get_predicate()
                        if left_predicate == "Top":
                            flag = False
                            right_atom = grounded_literal.right_atom
                            for _, term in enumerate(right_atom.get_entity()):
                                if term.type == "variable" and term.name not in context:
                                    flag = True
                                    break
                                elif term.type == "variable" and term.name in context:
                                    term.type = "constant"
                                    term.name = context[term.name]
                            if flag:
                                break
                            else:
                                tmp_predicate = right_atom.get_predicate()
                                tmp_entity = right_atom.get_entity()
                                if tmp_predicate in D and tmp_entity \
                                        in D[tmp_predicate]:
                                    tmp_D[tmp_predicate][tmp_entity] = D[tmp_predicate][tmp_entity]

                        elif right_predicate == "Top":
                            flag = False
                            left_atom = grounded_literal.left_atom
                            for _, term in enumerate(left_atom.get_entity()):
                                if term.type == "variable" and term.name not in context:
                                    flag = True
                                    break
                                elif term.type == "variable" and term.name in context:
                                    term.type = "constant"
                                    term.name = context[term.name]
                            if flag:
                                break
                            else:
                                tmp_predicate = left_atom.get_predicate()
                                tmp_entity = left_atom.get_entity()
                                if tmp_predicate in D and tmp_entity \
                                        in D[tmp_predicate]:
                                    tmp_D[tmp_predicate][tmp_entity] = D[tmp_predicate][tmp_entity]
                        else:
                            flag = False
                            right_atom = grounded_literal.right_atom
                            for _, term in enumerate(right_atom.get_entity()):
                                if term.type == "variable" and term.name not in context:
                                    flag = True
                                    break
                                elif term.type == "variable" and term.name in context:
                                    term.type = "constant"
                                    term.name = context[term.name]
                            if flag:
                                break
                            else:
                                tmp_predicate = right_atom.get_predicate()
                                tmp_entity = right_atom.get_entity()
                                if tmp_predicate in D and tmp_entity \
                                        in D[tmp_predicate]:
                                    tmp_D[tmp_predicate][tmp_entity] = D[tmp_predicate][tmp_entity]

                            left_atom = grounded_literal.left_atom
                            for _, term in enumerate(left_atom.get_entity()):
                                if term.type == "variable" and term.name not in context:
                                    flag = True
                                    break
                                elif term.type == "variable" and term.name in context:
                                    term.type = "constant"
                                    term.name = context[term.name]
                            if flag:
                                break
                            else:
                                tmp_predicate = left_atom.get_predicate()
                                tmp_entity = left_atom.get_entity()
                                if tmp_predicate in D and tmp_entity \
                                        in D[tmp_predicate]:
                                    tmp_D[tmp_predicate][tmp_entity] = D[tmp_predicate][tmp_entity]
                    else:
                        if grounded_literal.get_predicate() in ["Bottom", "Top"]:
                            raise ValueError("negative body atoms can not be Bottom or Top")

                        flag = False
                        for _, term in enumerate(grounded_literal.get_entity()):
                            if term.type == "variable" and term.name not in context:
                                flag = True
                                break
                            elif term.type == "variable" and term.name in context:
                                term.type = "constant"
                                term.name = context[term.name]
                        if flag:
                            break
                        else:
                            tmp_predicate = grounded_literal.get_predicate()
                            tmp_entity = grounded_literal.get_entity()
                            if tmp_predicate in D and tmp_entity \
                                    in D[tmp_predicate]:
                                tmp_D[tmp_predicate][tmp_entity] = D[tmp_predicate][tmp_entity]

                    t = apply(grounded_literal, tmp_D)
                    if len(t) == 0:
                        break
                    else:
                        n_T.append(t)

            if len(n_T) > 0 and len(n_T) == len(rule.negative_body_atoms):
                n_T = interval_merge(T)
            else:
                n_T = []

            if head_entity is not None:
                replaced_head_entity = []
                for term in head_entity:
                    if term.type == "constant":
                        replaced_head_entity.append(term)
                    else:
                        if term.name not in context:
                            raise ValueError("Head variables in Rule({}) do not appear in the body".format(str(rule)))
                        else:
                            new_term = Term.new_term(context[term.name])
                            replaced_head_entity.append(new_term)

                replaced_head_entity = tuple(replaced_head_entity)



            if len(T) == len(literals):
                T = interval_merge(T)
                exclude_t = []
                if len(T) != 0 and len(n_T) != 0:
                    exclude_t = interval_merge([T, n_T])
                if len(exclude_t) != 0:
                    T = Interval.diff(T, exclude_t)

                if len(T) != 0:
                    if not isinstance(rule.head, Atom):
                        tmp_D = defaultdict(lambda: defaultdict(list))
                        tmp_D[head_predicate][replaced_head_entity] = T
                        tmp_head = copy.deepcopy(rule.head)
                        tmp_head.set_entity(replaced_head_entity)
                        T = reverse_apply(tmp_head, tmp_D)

                    # store new facts to delta_new
                    diff_delta = []
                    if head_predicate not in D:
                        if head_predicate not in delta_new:
                            delta_new[head_predicate][replaced_head_entity] = T
                        else:
                            delta_new[head_predicate][replaced_head_entity] += T
                        if common_fragment is not None and common_fragment.cr_flag:
                            diff_delta = T

                    elif head_predicate in D and replaced_head_entity not in D[head_predicate]:
                        if head_predicate not in delta_new:
                            delta_new[head_predicate][replaced_head_entity] = T
                        else:
                            delta_new[head_predicate][replaced_head_entity] += T
                        if common_fragment is not None and common_fragment.cr_flag:
                            diff_delta = T
                    else:
                        coalesced_T_plus = coalescing(T + D[head_predicate][replaced_head_entity])
                        if not Interval.compare(coalesced_T_plus, D[head_predicate][replaced_head_entity]):
                            for interval1 in T:
                                if common_fragment is not None and common_fragment.cr_flag:
                                     diff_delta += Interval.diff(interval1, D[head_predicate][replaced_head_entity])
                                for interval2 in coalesced_T_plus:
                                    if Interval.intersection(interval1, interval2) is not None:
                                        flag = True
                                        for interval3 in D[head_predicate][replaced_head_entity]:
                                            if interval2 == interval3:
                                                flag = False
                                        if flag:
                                            delta_new[head_predicate][replaced_head_entity].append(interval2)

                    if common_fragment is not None and common_fragment.cr_flag and diff_delta is not None: # need to do judge whether it is overlap
                        for cr_interval in diff_delta:
                            if Interval.intersection(cr_interval, common_fragment.base_interval):
                                common_fragment.cr_flag = False
                                common_fragment.common = None
                                break
                            else:
                                if cr_interval.right_value <= common_fragment.base_interval.left_value:
                                    if cr_interval.right_value >= common_fragment.common.left_value:
                                        common_fragment.common.left_value = cr_interval.right_value
                                        common_fragment.common.left_open = not cr_interval.right_open
                                elif cr_interval.left_value >= common_fragment.base_interval.right_value:
                                    if cr_interval.left_value <= common_fragment.common.right_value:
                                        common_fragment.common.right_value = cr_interval.left_value
                                        common_fragment.common.right_open = not cr_interval.left_open
                                else:
                                    print(str(cr_interval))
                                    print(str(common_fragment.common))
                                    raise ValueError("Error Happen")

        elif global_literal_index == visited:
            ground_body(global_literal_index+1, visited, delta, context)

        else:
            current_literal = copy.deepcopy(literals[global_literal_index])
            if not isinstance(current_literal, BinaryLiteral):
                if current_literal.get_predicate() in ["Bottom", "Top"]:
                    ground_body(global_literal_index+1, delta, context)
                else:
                    for tmp_entity, tmp_interval, tmp_context in ground_unary_literal(current_literal, context, D, D_index):
                        tmp_entity_interval = tmp_interval if isinstance(tmp_interval, list) else list(tmp_interval)
                        tmp_delata = {global_literal_index: [tmp_entity, tmp_entity_interval]}
                        ground_body(global_literal_index+1,  visited, {**delta, **tmp_delata}, {**context, **tmp_context})

            else:
                left_predicate = current_literal.left_atom.get_predicate()
                right_predicate = current_literal.right_atom.get_predicate()

                if left_predicate in ["Bottom", "Top"]:
                    for tmp_entity, tmp_interval, tmp_context in ground_unary_literal(current_literal.right_atom,
                                                                                      context,  D, D_index):
                        tmp_entity_interval = tmp_interval if isinstance(tmp_interval, list) else list(tmp_interval)
                        tmp_delta = {global_literal_index: [tmp_entity, tmp_entity_interval]}
                        ground_body(global_literal_index + 1, visited, {**delta, **tmp_delta}, {**context, **tmp_context})

                elif right_predicate in ["Bottom", "Top"]:
                    for tmp_entity, tmp_interval, tmp_context in ground_unary_literal(current_literal.left_atom,
                                                                                      context, D, D_index):
                        tmp_entity_interval = tmp_interval if isinstance(tmp_interval, list) else list(tmp_interval)
                        tmp_delta = {global_literal_index: [tmp_entity, tmp_entity_interval]}
                        ground_body(global_literal_index + 1,visited, {**delta, **tmp_delta}, {**context, **tmp_context})

                else:
                    for left_entity, left_interval, tmp_context1 in ground_unary_literal(current_literal.left_atom, context, D):
                        for right_entity, right_interval, tmp_context2 in ground_unary_literal(current_literal.right_atom,
                                                                                               {**context,
                                                                                                **tmp_context1}, D):
                            tmp_left_interval = left_interval if isinstance(left_interval, list) else list(left_interval)
                            tmp_right_interval = right_interval if isinstance(right_interval, list) else list(right_interval)
                            tmp_delta = {global_literal_index: [left_entity, tmp_left_interval, right_entity, tmp_right_interval]}
                            ground_body(global_literal_index + 1, visited, {**delta, **tmp_delta}, {**context, **tmp_context1, **tmp_context2})

    for predicate in delta_old:
        for entity in delta_old[predicate]:
            for i in range(len(literals)):
                if literals[i].get_predicate() == predicate:
                    context = dict()
                    for term1, term2 in zip(literals[i].get_entity(), entity):
                        if term1.type == "constant" and term1.name != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                            break
                        else:
                            if term1.type == "variable":
                                context[term1.name] = term2.name
                    ground_body(0, i,  {i: [entity, delta_old[predicate][entity]]}, context)




if __name__ == "__main__":
    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"] = [Interval(2,8, False, False)]

    Delta = defaultdict(lambda: defaultdict(list))
    head = Atom("C")
    body = [Literal(Atom("A", tuple([Term("X", "variable")])), [Operator("Boxminus", Interval(1, 2, False, False))]), Literal(Atom("B"), [Operator("Diamondminus", Interval(0, 1, False, False))])]

    new_literal = BinaryLiteral(Atom("A", tuple([Term("X", "variable")])), Atom("B"), Operator("Since", Interval(1, 2, False, False)))
    body.append(new_literal)

    rule = Rule(head, body)
    seminaive_join(rule, Delta, D)

    for predicate in Delta:
        print(predicate)
        for interval in Delta[predicate]:
            print(interval)




import re, copy, random
from datetime import datetime
from meteor_reasoner.classes.atom import *
from meteor_reasoner.classes.term import *
from meteor_reasoner.classes.literal import *
from meteor_reasoner.classes.interval import *
from meteor_reasoner.classes.rule import *
from decimal import Decimal
ATOM_PATTERN = "(.*)\((.*)\)"
FACT_WITH_ENTITY_PATTERN = "(.*)\((.*)\)@(.*)"
FACT_WITHOUT_ENTITY_PATTERN = "(.*)@(.*)"
INTERVAL_TWO_POINTS_PATTERN = "(^[\(,\[])(-?\d+\.?\d*|-inf),(-?\d+\.?\d*|\+?inf)([\),\]])$"
INTERVAL_ONE_POINT_PATTERN =  "(^[\(,\[])(-?\d+\.?\d*|\+?inf)([\),\]])$"
OPERATOR_TWO_POINTS_PATTERN = "(.*)([\(,\[])(-?\d+\.?\d*|-inf),(-?\d+\.?\d*|\+?inf)([\),\]])$"
OPERATOR_ONE_POINT_PATTERN =  "(.*)([\(,\[])(-?\d+\.?\d*|-inf)([\),\]])$"


def random_return_name():
    strname = "_TMP_" + datetime.now().strftime("%H%M%S") + str(random.choice([1,2,4,5,6]))
    return strname


def parse_rule(line):
    """
    Parse a string-formed rule.
    Args:
        rule (str):

    Returns:
        A  list of rule instance(s).
    """
    line = line.strip()
    line = line.replace(" ", "")
    head, body = line.strip().split(":-")
    if "||" in body:
       body, negative_body = body.split("||")
    else:
        negative_body = []
    try:
        head_atom = parse_literal(head)
    except:
        raise Exception("{} has an incorrect syntax!".format(head))

    literals = []
    for literal_str in parse_body(body):
        try:
           literal = parse_literal(literal_str)
        except:
            raise Exception("{} has an incorrect syntax!".format(literal_str))
        literals.append(literal)

    negative_body = []
    if len(negative_body) != 0:
        for literal_str in parse_body(negative_body):
            try:
               literal = parse_literal(literal_str)

            except:
                raise Exception("{} has an incorrect syntax!".format(literal_str))
            negative_body.append(literal)

    rule = Rule(head_atom, literals, negative_body=negative_body)
    return rule


def parse_str_fact(line):
    result = re.search(FACT_WITH_ENTITY_PATTERN, line)
    if result:
        predicate = result.group(1)
        entity = tuple([Term(item) for item in result.group(2).split(",")])
        span = result.group(3)
        b = re.search(INTERVAL_TWO_POINTS_PATTERN, span)
        if b is None:  # contain one point
            try:
                Decimal(span)
                return predicate, entity, Interval(Decimal(span), Decimal(span), False, False)
            except ValueError:
                b = re.search(INTERVAL_ONE_POINT_PATTERN, span)
                if b is None:
                    raise ValueError("The interval can not be parsed correctly!")

                left_value = Decimal(b.group(2))
                right_value = Decimal(b.group(2))
                left_open, right_open = False, False

                if b.group(1) == "(":
                    left_open = True

                if b.group(3) == ")":
                    right_open = True

                if not Interval.is_valid_interval(left_value, right_value, left_open, right_open):
                    raise ValueError("{} contain an invalid interval!".format(line))

                return predicate, entity, Interval(left_value, right_value, left_open, right_open)

        else:
            left_value = Decimal(b.group(2))
            right_value = Decimal(b.group(3))
            left_open, right_open = False, False

            if b.group(1) == "(":
                left_open = True

            if b.group(4) == ")":
                right_open = True

            if not Interval.is_valid_interval(left_value, right_value, left_open, right_open):
                raise ValueError("{} contain an invalid interval!".format(line))

            return predicate, entity, Interval(left_value, right_value, left_open, right_open)

    else:
        result = re.search(FACT_WITHOUT_ENTITY_PATTERN, line)
        if result:
            predicate = result.group(1)
            entity = tuple([Term("nan")])
            span = result.group(2)
            b = re.search(INTERVAL_TWO_POINTS_PATTERN, span)
            if b is None:  # contain one point
                try:
                    Decimal(span)
                    return predicate, entity, Interval(Decimal(span), Decimal(span), False, False)
                except ValueError:
                    b = re.search(INTERVAL_ONE_POINT_PATTERN, span)
                    if b is None:
                        raise ValueError("The interval({}) can not be parsed correctly!".format(span))

                    left_value = Decimal(b.group(2))
                    right_value = Decimal(b.group(2))
                    left_open, right_open = False, False

                    if b.group(1) == "(":
                        left_open = True

                    if b.group(3) == ")":
                        right_open = True

                    if not Interval.is_valid_interval(left_value, right_value, left_open, right_open):
                        raise ValueError("{} contain an invalid interval!".format(line))

                    return predicate, entity, Interval(left_value, right_value, left_open, right_open)

            else:
                left_value = Decimal(b.group(2))
                right_value = Decimal(b.group(3))
                left_open, right_open = False, False

                if b.group(1) == "(":
                    left_open = True

                if b.group(4) == ")":
                    right_open = True

                if not Interval.is_valid_interval(left_value, right_value, left_open, right_open):
                    raise ValueError("{} contain an invalid interval!".format(line))

                return predicate, entity, Interval(left_value, right_value, left_open, right_open)
        else:
            raise ValueError("{} is an invalid fact!".format(line))


def parse_atom(atom_str):
    """
    Parse a string-form atom into an atom instance.
    Args:
        atom_str (str): a string-form atom, e.g.  A(X,y). In particular, we specify that
        a variable starts with the first letter capitalized.

    Returns:
        an atom instance.
    """

    head_result = re.search(ATOM_PATTERN, atom_str)
    if head_result is not None:
        head_predicate = head_result.group(1)
        head_entity = head_result.group(2).split(",")
        entity_list = []
        for item in head_entity:
            if item[0].isupper():  # is a variable
                entity_list.append(Term(item, "variable"))
            else:
                entity_list.append(Term(item, "constant"))

        atom = Atom(head_predicate, tuple(entity_list))
    else:
        atom = Atom(atom_str, tuple([Term("nan")]))

    return atom


def parse_body(body):
    """
    Split literals separated by comma in the body into a list.

    Args:
        body (str): a string consisting of one or multiple literals separated by comma

    Returns:
        a list of strings, each of which represents one literal

    """
    literals = []
    left = False
    current_str = ""
    for item in body:
        if item in ["(", "["]:
            left = True

        if item != ",":
            if item in ["(", "["]:
                left = True
            if item in [")", "]"]:
                left = False
            current_str += item
        else:
            if not left:
                literals.append(current_str)
                current_str = ""
            else:
                current_str += item
    literals.append(current_str)
    return literals


def parse_operator(operator_str):
    """
    Parse a string-form operator into a operator instance.

    Args:
        operator_str (str): a string-form operator, e.g. Boxminus[2,3]

    Returns:
       Operator instance.
    """
    b = re.search(OPERATOR_TWO_POINTS_PATTERN, operator_str)
    if b is None:
        b = re.search(OPERATOR_ONE_POINT_PATTERN, operator_str)
        if b is None:
            raise ValueError("The interval can not be parsed correctly!")

        left_value = Decimal(b.group(3))
        right_value = Decimal(b.group(3))
        left_open, right_open = False, False

        if b.group(2) == "(":
            left_open = True

        if b.group(4) == ")":
            right_open = True

        if b.group(1) == "SOMETIME" and left_value >= 0:
            return Operator("Diamondplus", Interval(left_value, right_value, left_open, right_open))
        elif b.group(1) == "SOMETIME" and left_value < 0:
            return Operator("Diamondminus", Interval(-right_value, -left_value, right_open, left_open))
        elif b.group(1) == "ALWAYS" and left_value >= 0:
            return Operator("Boxplus", Interval(left_value, right_value, left_open, right_open))
        elif b.group(1) == "ALWAYS" and left_value <= 0:
            return Operator("Boxminus", Interval(-right_value, -left_value, right_open, left_open))
        elif b.group(1) == "UNTIL" and left_value < 0:
            return Operator("Since", Interval(-right_value, -left_value, right_open, left_open))
        elif b.group(1) == "UNTIL" and left_value >=0:
            return Operator("Until", Interval(left_value, right_value, left_open, right_open))
        else:
            return Operator(b.group(1), Interval(left_value, right_value, left_open, right_open))

    left_value = Decimal(b.group(3))
    right_value = Decimal(b.group(4))
    left_open, right_open = False, False

    if b.group(2) == "(":
        left_open = True

    if b.group(5) == ")":
        right_open = True

    if b.group(1) == "SOMETIME" and left_value>=0:
        return Operator("Diamondplus", Interval(left_value, right_value, left_open, right_open))
    elif b.group(1) == "SOMETIME" and left_value<0:
        return Operator("Diamondminus", Interval(-right_value,-left_value, right_open, left_open))
    elif b.group(1) == "ALWAYS" and left_value>=0:
        return Operator("Boxplus", Interval(left_value, right_value, left_open, right_open))
    elif b.group(1) == "ALWAYS" and left_value < 0:
        return Operator("Boxminus", Interval(-right_value, -left_value, right_open, left_open))
    elif b.group(1) == "UNTIL" and left_value < 0:
        return Operator("Since", Interval(-right_value, -left_value, right_open, left_open))
    elif b.group(1) == "UNTIL" and left_value >= 0:
        return Operator("Until", Interval(left_value, right_value, left_open, right_open))
    else:
       return Operator(b.group(1), Interval(left_value, right_value, left_open, right_open))


def parse_literal(literal):
    """
    We write different regex expressions to parse the string-form literal
    Args:
        literal (str):

    Returns:
        BinaryLiteral instance or Literal instance.

    """

    pattern1 = "Boxminus[\[,\(]-?\d+\.?\d*[\),\]]"
    pattern2 = "Boxplus[\[,\(]-?\d+\.?\d*[\),\]]"
    pattern3 = "Diamondminus[\[,\(]-?\d+\.?\d*[\),\]]"
    pattern4 = "Diamondplus[\[,\(]-?\d+\.?\d*[\),\]]"
    pattern4_1 = "SOMETIME[\[,\(]-?\d+\.?\d*[\),\]]"
    pattern4_2 = "ALWAYS[\[,\(]-?\d+\.?\d*[\),\]]"



    pattern5 = "Boxminus[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"
    pattern6 = "Boxplus[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"
    pattern7 = "Diamondminus[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"
    pattern8 = "Diamondplus[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"
    pattern8_1 = "SOMETIME[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"
    pattern8_2 = "ALWAYS[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"

    pattern9 =  "Boxminus[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    pattern10 = "Boxplus[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    pattern11 = "Diamondminus[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    pattern12 = "Diamondplus[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    pattern12_1 = "SOMETIME[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    pattern12_2 = "ALWAYS[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"


    pattern13 = "Boxminus[\[,\(]-inf,-?\d+\.?\d*[\),\]]"
    pattern14 = "Boxplus[\[,\(]-inf,-?\d+\.?\d*[\),\]]"
    pattern15 = "Diamondminus[\[,\(]-inf,-?\d+\.?\d*[\),\]]"
    pattern16 = "Diamondplus[\[,\(]-inf,-?\d+\.?\d*[\),\]]"
    pattern16_1 = "SOMETIME[\[,\(]-inf,-?\d+\.?\d*[\),\]]"
    pattern16_2 = "ALWAYS[\[,\(]-inf,-?\d+\.?\d*[\),\]]"

    since_pattern1 = "Since[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"
    since_pattern3 = "Since[\[,\(]-?\d+\.?\d*[\),\]]"
    until_pattern2 = "Until[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"
    unitl_pattern4 = "Until[\[,\(]-?\d+\.?\d*[\),\]]"
    UNTIL_pattern5 = "UNTIL[\[,\(]-?\d+\.?\d*,-?\d+\.?\d*[\),\]]"
    UNTIL_pattern6 = "UNTIL[\[,\(]-?\d+\.?\d*[\),\]]"

    since_pattern00 = "Since[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    since_pattern01 = "Since[\[,\(]-inf,-?\d+\.?\d*[\),\]]"
    UNTIL_pattern02 = "UNTIL[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    UNTIL_pattern03 = "UNTIL[\[,\(]-inf,-?\d+\.?\d*[\),\]]"

    until_pattern11 = "Until[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    until_pattern12 = "Until[\[,\(]-inf,-?\d+\.?\d*[\),\]]"
    UNTIL_pattern13 = "UNTIL[\[,\(]-?\d+\.?\d*,\+?inf[\),\]]"
    UNTIL_pattern14 = "UNTIL[\[,\(]-inf,-?\d+\.?\d*[\),\]]"


    # check whether it is a since or a until literal
    result = re.findall("|".join([since_pattern00, since_pattern01, until_pattern11, until_pattern12, UNTIL_pattern02, UNTIL_pattern03,
                                  UNTIL_pattern13, UNTIL_pattern14, since_pattern1,
                                  since_pattern3, until_pattern2, unitl_pattern4, UNTIL_pattern5, UNTIL_pattern6]), literal)

    if len(result) != 0:
        begin_index = literal.index(result[0])
        left_literal = literal[:begin_index]
        right_literal = literal[begin_index + len(result[0]):]
        left_literal = parse_literal(left_literal)
        right_literal = parse_literal(right_literal)

        op = parse_operator(result[0])
        b_literal = BinaryLiteral(left_literal, right_literal, op)

        return b_literal

    else:
        # check whether it is a literal with temporal operator(s)
        #result = re.findall("|".join([pattern12_2]), literal)
        result = re.findall("|".join([pattern9, pattern10, pattern11, pattern12, pattern12_1, pattern12_2, pattern13, pattern14,
                                      pattern15, pattern16, pattern16_1, pattern16_2, pattern5, pattern6, pattern7, pattern8,  pattern8_1, pattern8_2, pattern1,
                                      pattern2, pattern3, pattern4, pattern4_1, pattern4_2]), literal)
        if len(result) != 0:
            operators = []
            for operator_str in result:
                operator_ins = parse_operator(operator_str)
                operators.append(operator_ins)

            atom_str = literal[len("".join(result)):]
            atom = parse_atom(atom_str)
            literal = Literal(atom, operators)
            return literal

        else:
            # literal without any temporal operators
            atom = parse_atom(literal)
            t_literal = Literal(atom, [])
            return t_literal


if __name__ == "__main__":
    operator_str = "ALWAYS[-1,-3]A"
    lit = parse_literal(operator_str)
    print(lit)



from meteor_reasoner.graphutil.graph import *


def no_new_facts(delta_new, D, limit):
    for predicate in delta_new:
        for entity in delta_new[predicate]:
            if predicate in D and entity in D[predicate]:
                diff_interval_list = Interval.diff_list(delta_new[predicate][entity], D[predicate][entity])
                if len(diff_interval_list) > 0:
                    for interval in diff_interval_list:
                        new_interval = Interval.intersection(limit, interval)
                        if new_interval is not None:
                            return False
            else:
                for interval in delta_new[predicate][entity]:
                    new_interval = Interval.intersection(limit, interval)
                    if new_interval is not None:
                       return False
    return True


def literal_contain_no_variable(literal):
    if isinstance(literal, Atom):
        if literal.get_predicate() not in ["Bottom", "Top"] and len(literal.entity) == 1 and literal.entity[0].name == "nan":
            return True
    elif isinstance(literal, Literal):
        if isinstance(literal, Literal) and len(literal.atom.entity) == 1 and literal.atom.entity[0].name == "nan":
            return True
    elif isinstance(literal, BinaryLiteral):
        if literal.left_atom.get_predicate() in ["Bottom", "Top"] or (
                len(literal.left_atom.entity) == 1 and literal.left_atom.entity[0].name == "nan"):
            if literal.right_atom.get_predicate() in ["Bottom", "Top"] or (
                    len(literal.right_atom.entity) == 1 and literal.right_atom.entity[0].name == "nan"):
                return True


def split_rules_predicates(program):
    """
    This function split all rules in the given program into non-recursive rules and recursive rules.
    Besides, this function also split all predicates in the given program into involved predicates and automata predicated.

    None-recursive rules: rules needed to be completely materialized before running the automata
    Recursive rules: rules that can not be completely materialized
    Involved predicates refer to invariable predicates but are needed to be considered in the Automata construction
    Automata predicates refer to variable predicates in the Automata construction whose states are uncertain
    Args:
        program (list of Rule instances):

    Returns:
        set, set, set, set

    """
    graph = Graph(program)

    non_recursive_rules, automata_rules = graph.split_rules("a1:FullProfessor")
    involved_predicates = set()
    automata_predicates = set()

    for rule in non_recursive_rules:
        involved_predicates.add(rule.head.get_predicate())
        for literal in rule.body:
            if isinstance(literal, BinaryLiteral):
                if literal.left_atom.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.left_atom.get_predicate())
                if literal.right_atom.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.right_atom.get_predicate())
            else:
                if literal.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.get_predicate())

    for rule in automata_rules:
        automata_predicates.add(rule.head.get_predicate())
        involved_predicates.add(rule.head.get_predicate())
        for literal in rule.body:
            if isinstance(literal, BinaryLiteral):
                if literal.left_atom.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.left_atom.get_predicate())
                    automata_predicates.add(rule.head.get_predicate())
                if literal.right_atom.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.right_atom.get_predicate())
                    automata_predicates.add(rule.head.get_predicate())
            else:
                if literal.get_predicate() not in ["Bottom", "Top"]:
                    involved_predicates.add(literal.get_predicate())
                    automata_predicates.add(rule.head.get_predicate())

    return (non_recursive_rules, automata_rules, involved_predicates, automata_predicates)



from meteor_reasoner.utils.loader import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.utils.operate_dataset import *
from meteor_reasoner.materialization.materialize import materialize
from meteor_reasoner.utils.ruler_interval import *

class CanonicalRepresentation:
    def __init__(self, D, Program):
        self.D = D
        self.D_index = defaultdict(lambda : defaultdict(list))
        self.Program = Program
        self.initilization()

    def initilization(self):
        coalescing_d(self.D)
        build_index(self.D,  self.D_index )
        self.points, self.min_x, self.max_x = get_dataset_points_x(self.D, min_x_flag=True)
        self.z, self.gcd = get_gcd(self.Program)
        _, self.initial_ruler_intervals = get_initial_ruler_intervals(self.points, left_border= self.min_x,
                                                                      right_border=self.max_x, gcd=self.gcd)
        self.left_dict, self.right_dict = construct_left_right_pattern(self.points, self.gcd)


def find_common_fragment(D1, D2, rules, varrho):
    points, min_x, max_x = get_dataset_points_x(D2, min_x_flag=True)
    _, gcd = get_gcd(rules)
    _, initial_window_ruler_intervals = get_initial_ruler_intervals(points, left_border=min_x, right_border=max_x, gcd=gcd)
    left_point = Interval(varrho.left_value, varrho.left_value, False, False)
    right_point = Interval(varrho.right_value, varrho.right_value, False, False)
    left_i = initial_window_ruler_intervals.index(left_point)
    right_i = initial_window_ruler_intervals.index(right_point)

    # check whether in the interval [t_D^{-}, t_D^{+}], D' and D'' are the same
    middle_i = left_i
    while middle_i <= right_i:
        flag = False
        ruler_interval = initial_window_ruler_intervals[middle_i]
        for predicate in D2:
            for entity in D2[predicate]:
                if interval_intesection_intervallist(ruler_interval, D2[predicate][entity]):
                    if predicate not in D1 or entity not in D1[predicate] or not \
                            interval_intesection_intervallist(ruler_interval, D1[predicate][entity]):
                        flag = True
                        break
            if flag:
                return None, None
        middle_i += 1

    left_border = None
    while left_i > 0:
        left_i -= 1
        ruler_interval = initial_window_ruler_intervals[left_i]
        for predicate in D2:
            flag = False
            for entity in D2[predicate]:
                if interval_intesection_intervallist(ruler_interval, D2[predicate][entity]):
                    if predicate not in D1 or entity not in D1[predicate] or not \
                            interval_intesection_intervallist(ruler_interval, D1[predicate][entity]):
                        left_border = ruler_interval
                        flag = True
                        break
            if flag:
                break

    right_border = None
    while right_i < len(initial_window_ruler_intervals)-1:
        right_i += 1
        ruler_interval = initial_window_ruler_intervals[right_i]
        flag = False
        for predicate in D2:
            for entity in D2[predicate]:
                if interval_inclusion_intervallist(ruler_interval, D2[predicate][entity]):
                    if predicate in D1 and entity in D1[predicate] and interval_inclusion_intervallist(
                            ruler_interval, D1[predicate][entity]):
                        continue
                    else:
                        if ruler_interval.left_open:
                            right_border = Interval(ruler_interval.left_value, ruler_interval.left_value, False, False)
                        else:
                            right_border = Interval(ruler_interval.left_value, ruler_interval.left_value, True, True)
                        flag = True
                        break
            if flag:
                break
        if flag:
            break

    if left_border is None:
        left_value, left_open = decimal.Decimal("-inf"), True
    else:
        left_value, left_open = left_border.left_value, left_border.left_open

    if right_border is None:
        right_value, right_open = decimal.Decimal("inf"), True
    else:
        right_value, right_open = right_border.left_value, right_border.left_open

    varrho_left_range = None
    if Interval.is_valid_interval(left_value, varrho.left_value, left_open, True):
        varrho_left_range = Interval(left_value, varrho.left_value, left_open, True)
    varrho_right_range = None
    if Interval.is_valid_interval(varrho.right_value, right_value, True, right_open):
        varrho_right_range = Interval(varrho.right_value, right_value, True, right_open)

    return varrho_left_range, varrho_right_range


def has_same_pattern(ruler_intervals1, ruler_intervals2):
    """
    Check whether the give two ruler interval lists have the same pattern
    Args:
        ruler_intervals1: a list of ruler-intervals
        ruler_intervals2: a list of ruler-intervals

    Returns:
        True or False
    """
    if len(ruler_intervals1) != len(ruler_intervals2):
        return False
    if ruler_intervals1[0].left_open != ruler_intervals2[0].left_open or ruler_intervals1[-1].right_open != ruler_intervals2[-1].right_open:
        return False
    for ruler1, ruler2 in zip(ruler_intervals1, ruler_intervals2):
        if abs(ruler1.right_value-ruler1.left_value) != abs(ruler2.right_value - ruler2.left_value):
            return False
    return True


def has_same_facts(ruler_intervals1, ruler_intervals2, D):
    """
    Check whether the two same-pattern ruler lists have the same facts at each corresponding ruler-interval
    Args:
        ruler_intervals1: a list of ruler-intervals
        ruler_intervals2: a list of ruler-intervals
        D: contain all relational facts
    Returns:
        True or False
    """

    for ruler1, ruler2 in zip(ruler_intervals1, ruler_intervals2):
        for predicate in D:
            for entity in D[predicate]:
                if interval_inclusion_intervallist(ruler1, D[predicate][entity]) and \
                        not interval_inclusion_intervallist(ruler2, D[predicate][entity]):
                    return False
    return True


def find_left_period(D, left_interval_range, CR, w):
    if left_interval_range.left_value != decimal.Decimal("-inf"):
        big_ruler_intervals = CR.initial_ruler_intervals[:]
        starting_ruler_interval = None
        while big_ruler_intervals[0].left_value >= left_interval_range.left_value:
            ruler_interval_len = []
            cnt = 0
            if big_ruler_intervals[0].left_open:
                new_ruler_interval = Interval(big_ruler_intervals[0].left_value, big_ruler_intervals[0].left_value,
                                              False, False)
                if starting_ruler_interval is None:
                    starting_ruler_interval = new_ruler_interval
                if new_ruler_interval.left_value > left_interval_range.left_value:
                   big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
                   continue
                elif new_ruler_interval.left_value == left_interval_range.left_value and new_ruler_interval.left_open == left_interval_range.left_open:
                    big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
                    continue
                else:
                    break

            for i in range(len(big_ruler_intervals)):
                if big_ruler_intervals[i].left_open:
                    ruler_interval_len.append(
                        str(abs(big_ruler_intervals[i].right_value - big_ruler_intervals[i].left_value)))
                    cnt += 1
                if cnt >= len(CR.left_dict):
                    break
            tmp_pattern = "#".join(ruler_interval_len)
            current_the_most_left_ruler_interval = big_ruler_intervals[0]
            next_ruler_interval_len = CR.left_dict[tmp_pattern]
            new_ruler_interval = (Interval(current_the_most_left_ruler_interval.left_value - next_ruler_interval_len,
                                           current_the_most_left_ruler_interval.left_value, True, True))
            if new_ruler_interval.left_value > left_interval_range.left_value:
                big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
                continue
            elif new_ruler_interval.left_value == left_interval_range.left_value and new_ruler_interval.left_open == left_interval_range.left_open:
                big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
                continue
            else:
                break

        big_ruler_intervals = big_ruler_intervals[0: big_ruler_intervals.index(starting_ruler_interval) + 1]

        len_big_ruler_intervals = len(big_ruler_intervals)
        for i in range(len_big_ruler_intervals - 1, -1, -1):
            if big_ruler_intervals[i].left_open:
                continue
            j = i
            first_interval = []
            tmp_w = w
            first_flag = False
            while tmp_w >= 0:
                next_ruler_interval_len = abs(big_ruler_intervals[j].right_value - big_ruler_intervals[j].left_value)
                tmp_w -= next_ruler_interval_len
                if tmp_w >= 0:
                    first_interval = [big_ruler_intervals[j]] + first_interval
                j -= 1
                if j < 0:
                    first_flag = True
                    break
            if first_flag:
                break

            for k in range(i - 1, -1, -1):
                if big_ruler_intervals[k].left_open:
                    continue
                j = k
                second_interval = []
                tmp_w = w
                second_flag = False
                while tmp_w >= 0:
                    next_ruler_interval_len = abs(
                        big_ruler_intervals[j].left_value - big_ruler_intervals[j].right_value)
                    tmp_w -= next_ruler_interval_len
                    if tmp_w >= 0:
                        second_interval = [big_ruler_intervals[j]] + second_interval
                    j -= 1
                    if j < 0:
                        second_flag = True
                        break
                if second_flag:
                    break

                if has_same_pattern(second_interval, first_interval) and has_same_facts(second_interval, first_interval,
                                                                                        D):
                    varrho_left_dict = defaultdict(list)
                    for ruler in first_interval:
                        for predicate in D:
                            for entity in D[predicate]:
                                if interval_inclusion_intervallist(ruler, D[predicate][entity]):
                                    varrho_left_dict[ruler].append(Atom(predicate, entity))
                    return first_interval, second_interval, varrho_left_dict

        return None, None, None

    else:
        threshold = 100
        big_ruler_intervals = CR.initial_ruler_intervals[:]
        starting_ruler_interval = None
        for _ in range(threshold+len(CR.initial_ruler_intervals)+1):
            ruler_interval_len = []
            cnt = 0
            if big_ruler_intervals[0].left_open:
                new_ruler_interval = Interval(big_ruler_intervals[0].left_value, big_ruler_intervals[0].left_value, False, False)
                if starting_ruler_interval is None:
                    starting_ruler_interval = new_ruler_interval
                big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals
                continue

            for i in range(len(big_ruler_intervals)):
                if big_ruler_intervals[i].left_open:
                    ruler_interval_len.append(str(abs(big_ruler_intervals[i].right_value - big_ruler_intervals[i].left_value)))
                    cnt += 1
                if cnt >= len(CR.left_dict):
                    break
            tmp_pattern = "#".join(ruler_interval_len)
            current_the_most_left_ruler_interval = big_ruler_intervals[0]
            next_ruler_interval_len = CR.left_dict[tmp_pattern]
            new_ruler_interval = (Interval(current_the_most_left_ruler_interval.left_value - next_ruler_interval_len,
                                           current_the_most_left_ruler_interval.left_value, True, True))
            big_ruler_intervals = [new_ruler_interval] + big_ruler_intervals

        big_ruler_intervals = big_ruler_intervals[0: big_ruler_intervals.index(starting_ruler_interval)+1]

        len_big_ruler_intervals = len(big_ruler_intervals)
        for i in range(len_big_ruler_intervals-1, -1, -1):
            if big_ruler_intervals[i].left_open:
                continue
            j = i
            first_interval = []
            tmp_w = w
            while tmp_w >= 0:
                next_ruler_interval_len = abs(big_ruler_intervals[j].right_value - big_ruler_intervals[j].left_value)
                tmp_w -= next_ruler_interval_len
                if tmp_w >= 0:
                    first_interval = [big_ruler_intervals[j]] + first_interval
                j -= 1

            for k in range(i - 1, -1, -1):
                if big_ruler_intervals[k].left_open:
                    continue
                j = k
                second_interval = []
                tmp_w = w
                while tmp_w >= 0:
                    next_ruler_interval_len = abs(big_ruler_intervals[j].left_value - big_ruler_intervals[j].right_value)
                    tmp_w -= next_ruler_interval_len
                    if tmp_w >= 0:
                        second_interval = [big_ruler_intervals[j]] + second_interval
                    j -= 1

                if has_same_pattern(second_interval, first_interval) and has_same_facts(second_interval, first_interval,D):
                     varrho_left_dict = defaultdict(list)
                     for ruler in first_interval:
                         for predicate in D:
                             for entity in D[predicate]:
                                 if interval_inclusion_intervallist(ruler, D[predicate][entity]):
                                     varrho_left_dict[ruler].append(Atom(predicate, entity))
                     return first_interval, second_interval,  varrho_left_dict


def find_right_period(D, right_interval_range, CR, w):
    if right_interval_range.right_value != decimal.Decimal("inf"):
        big_ruler_intervals = CR.initial_ruler_intervals[:]
        starting_ruler_interval = None
        while big_ruler_intervals[-1].right_value <= right_interval_range.right_value:
            ruler_interval_len = []
            cnt = 0
            if big_ruler_intervals[-1].left_open:
                new_ruler_interval = Interval(big_ruler_intervals[-1].right_value, big_ruler_intervals[-1].right_value,
                                              False, False)
                if starting_ruler_interval is None:
                    starting_ruler_interval = new_ruler_interval

                if new_ruler_interval.right_value < right_interval_range.right_value:
                   big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
                   continue
                elif new_ruler_interval.right_value == right_interval_range.right_value and new_ruler_interval.right_open == right_interval_range.right_open:
                    big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
                    continue
                else:
                    break

            for i in range(len(big_ruler_intervals)-1, -1, -1):
                if big_ruler_intervals[i].left_open:
                    ruler_interval_len.append(
                        str(abs(big_ruler_intervals[i].right_value - big_ruler_intervals[i].left_value)))
                    cnt += 1
                if cnt >= len(CR.right_dict):
                    break
            tmp_pattern = "#".join(ruler_interval_len)
            current_the_most_right_ruler_interval = big_ruler_intervals[-1]
            next_ruler_interval_len = CR.right_dict[tmp_pattern]
            new_ruler_interval = (Interval(current_the_most_right_ruler_interval.right_value,
                                           current_the_most_right_ruler_interval.right_value + next_ruler_interval_len, True, True))
            if new_ruler_interval.right_value < right_interval_range.right_value:
                big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
            elif new_ruler_interval.right_value == right_interval_range.right_value and new_ruler_interval.right_open == right_interval_range.right_open:
                big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
            elif new_ruler_interval.right_value == right_interval_range.right_value and not right_interval_range.right_open:
                big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
            else:
                break
        big_ruler_intervals = big_ruler_intervals[big_ruler_intervals.index(starting_ruler_interval):]

        len_big_ruler_intervals = len(big_ruler_intervals)
        for i in range(0, len_big_ruler_intervals-1):
            if big_ruler_intervals[i].left_open:
                continue
            j = i
            first_interval = []
            tmp_w = w
            first_flag = False
            while tmp_w >= 0:
                next_ruler_interval_len = abs(big_ruler_intervals[j].right_value - big_ruler_intervals[j].left_value)
                tmp_w -= next_ruler_interval_len
                if tmp_w >= 0:
                    first_interval =first_interval + [big_ruler_intervals[j]]
                j += 1
                if j >= len(big_ruler_intervals):
                    first_flag = True
                    break
            if first_flag:
                break

            for k in range(i+1,len_big_ruler_intervals):
                if big_ruler_intervals[k].left_open:
                    continue
                j = k
                second_interval = []
                tmp_w = w
                second_flag = False
                while tmp_w >= 0:
                    next_ruler_interval_len = abs(
                        big_ruler_intervals[j].left_value - big_ruler_intervals[j].right_value)
                    tmp_w -= next_ruler_interval_len
                    if tmp_w >= 0:
                        second_interval = second_interval + [big_ruler_intervals[j]]
                    j += 1
                    if j >= len(big_ruler_intervals):
                        second_flag = True
                        break
                if second_flag:
                    break

                if has_same_pattern(second_interval, first_interval) and has_same_facts(second_interval, first_interval,D):
                    varrho_right_dict = defaultdict(list)
                    for ruler in first_interval:
                        for predicate in D:
                            for entity in D[predicate]:
                                if interval_inclusion_intervallist(ruler, D[predicate][entity]):
                                    varrho_right_dict[ruler].append(Atom(predicate, entity))
                    return first_interval, second_interval,  varrho_right_dict
        return None, None, None
    else:
        threshold = 100
        big_ruler_intervals = CR.initial_ruler_intervals[:]
        starting_ruler_interval = None
        for _ in range(threshold + len(CR.initial_ruler_intervals) + 1):
            ruler_interval_len = []
            cnt = 0
            if big_ruler_intervals[-1].left_open:
                new_ruler_interval = Interval(big_ruler_intervals[-1].right_value, big_ruler_intervals[-1].right_value,
                                              False, False)
                if starting_ruler_interval is None:
                    starting_ruler_interval = new_ruler_interval
                if new_ruler_interval.right_value < right_interval_range.right_value:
                    big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
                elif new_ruler_interval.right_value == right_interval_range.right_value and new_ruler_interval.right_open == right_interval_range.right_open:
                    big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
                continue

            for i in range(len(big_ruler_intervals)-1, -1, -1):
                if big_ruler_intervals[i].left_open:
                    ruler_interval_len.append(
                        str(abs(big_ruler_intervals[i].right_value - big_ruler_intervals[i].left_value)))
                    cnt += 1
                if cnt >= len(CR.left_dict):
                    break
            tmp_pattern = "#".join(ruler_interval_len)
            current_the_most_right_ruler_interval = big_ruler_intervals[-1]
            next_ruler_interval_len = CR.left_dict[tmp_pattern]
            new_ruler_interval = (Interval(current_the_most_right_ruler_interval.right_value,
                                           current_the_most_right_ruler_interval.right_value + next_ruler_interval_len, True, True))
            if new_ruler_interval.right_value < right_interval_range.right_value:
                big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]
            elif new_ruler_interval.right_value == right_interval_range.right_value and new_ruler_interval.right_open == right_interval_range.right_open:
                big_ruler_intervals = big_ruler_intervals + [new_ruler_interval]

        big_ruler_intervals = big_ruler_intervals[big_ruler_intervals.index(starting_ruler_interval):]

        len_big_ruler_intervals = len(big_ruler_intervals)
        for i in range(0, len_big_ruler_intervals - 1):
            if big_ruler_intervals[i].left_open:
                continue
            j = i
            first_interval = []
            tmp_w = w
            while tmp_w >= 0:
                next_ruler_interval_len = abs(big_ruler_intervals[j].right_value - big_ruler_intervals[j].left_value)
                tmp_w -= next_ruler_interval_len
                if tmp_w >= 0:
                    first_interval = first_interval + [big_ruler_intervals[j]]
                j += 1

            for k in range(i + 1, len_big_ruler_intervals):
                if big_ruler_intervals[k].left_open:
                    continue
                j = k
                second_interval = []
                tmp_w = w
                while tmp_w >= 0:
                    next_ruler_interval_len = abs(
                        big_ruler_intervals[j].left_value - big_ruler_intervals[j].right_value)
                    tmp_w -= next_ruler_interval_len
                    if tmp_w >= 0:
                        second_interval = second_interval + [big_ruler_intervals[j]]
                    j -= 1

                if has_same_pattern(second_interval, first_interval) and has_same_facts(second_interval, first_interval, D):
                    varrho_right_dict = defaultdict(list)
                    for ruler in first_interval:
                        for predicate in D:
                            for entity in D[predicate]:
                                if interval_inclusion_intervallist(ruler, D[predicate][entity]):
                                    varrho_right_dict[ruler].append(Atom(predicate, entity))
                    return first_interval, second_interval, varrho_right_dict


def main(CR, w):
    while True:
        D1 = copy.deepcopy(CR.D)
        materialize(CR.D, rules=CR.Program, K=1)
        D2 = copy.deepcopy(CR.D)
        varrho_left_range, varrho_right_range = find_common_fragment(D1, D2, CR.Program, Interval(CR.min_x, CR.max_x, False, False))
        if varrho_left_range is None or varrho_right_range is None:
            continue
        varrho_left_1, varrho_left_2, varrho_left_dict = find_left_period(D1, varrho_left_range, CR, w)
        if varrho_left_1 is not None:
            varrho_right_1, varrho_right_2, varrho_right_dict = find_right_period(D1, varrho_right_range, CR, w)
            if varrho_right_1 is not None:
                return D1, varrho_left_1, varrho_left_2, varrho_left_dict,  varrho_right_1, varrho_right_2, varrho_right_dict


if __name__ == "__main__":
    raw_data = ["P(mike,mike)@[0,1]"]
    raw_program = ["ALWAYS[0,1]P(X,X):- P(X,X)"]
    w = 1

    raw_data = [
        "P(a,a)@(0,5)",
        "P(a,a)@6",
        "Q(b)@[1,5]",
        "Q(a)@(1,4)",
        "R(a,b)@(0,4)",
        "R(a,b)@(4,7]", "R(b,b)@1",
        "S(a)@0", "S(a)@1", "S(b)@[1,5)"
    ]

    raw_program = [
        "P(X,Y):- Boxminus[0,1]Q(Y), Diamondplus[1,2]S(X)",
        "Boxplus[2,3]R(X,X):-P(X,Y), S(Y)",
        "S(X):-Diamondminus[1,2]Q(y), Diamondminus[3,4]S(X)"
    ]
    w = 8

    raw_data = [
        "P(a,a)@[0,4)", "P(a,a)@(4,5]", "P(a,b)@(3,6]",
        "Q(b)@1", "Q(a)@2",
        "R(a,b)@(0,5]", "R(b,b)@0", "R(b,b)@[5,7)",
        "S(a)@[0,3)", "S(a)@(3,7)", "S(b)@0"
    ]
    raw_program = [
        "P(X,Y):-Boxminus(0,1]Q(Y), Diamondplus[1,2)R(X)",
        "Boxminus(1,2)R(X,X):- P(X,Y), S(Y)",
        "S(X,Y):- Diamondminus(1,2)Q(Y), Diamondminu(2,4]S(X)"
    ]
    w = 8

    raw_data = [
        "P(a,a)@(-2,3]", "P(a,a)@4", "P(a,b)@(-1,3)",
        "A(b)@[-1,4)", "Q(a)@(4,7)", "Q(a)@7",
        "R(a,a)@0", "R(a,b)@[-2,3)", "R(b,b)@(-1,3]",
        "S(a)@2", "S(a)@(-2,5]", "S(b)@[0,5)"
    ]

    raw_program = [
        "P(X,Y):- Boxminus[2]Q(Y), Diamondplus[2]S(X)",
        "Boxminus(1,2)R(X,X):- Diamondplus[1]P(X,Y), S(Y)",
        "S(X):- Diamondminus(1,2)Q(Y), Diamondminus[2]S(X)"
    ]
    w = 6

    raw_data = [
        "P(a,a)@0", "P(a,a)@6", "P(b,b)@(0,6)", "P(a,b)@(0,6)",
        "Q(b)@(-1,3)", "Q(a)@4", "Q(a)@5",
        "R(a,b)@(0,4)", "R(a,b)@(-4,0)", "R(b,b)@(0,4)", "R(b,b)@(4,5)",
        "S(a)@(-2,2)", "S(a)@3", "S(b)@[2,3]"
    ]
    raw_program = [
        "P(X,Y):- Boxplus[1,4]Q(Y), Diamondplus[2,3]S(X), Boxplus[3]P(X,Y)",
        "Boxminus[1]R(X,X):- Diamondplus[1]P(X,Y),Boxplus(1,2)@S(Y)",
        "S(X):- Diamondminus(1,2)Q(Y), Diamondminus[2]@S(X), Diamondminus(2,3)Q(Y)"
    ]
    w = 8

    D = load_dataset(raw_data)
    print("Dataset:")
    print_dataset(D)
    D_index = defaultdict(lambda: defaultdict(list))
    Program = load_program(raw_program)
    print("Program:")
    print([str(rule) for rule in Program])
    print("===========")
    CR = CanonicalRepresentation(D, Program)
    D, varrho_left1, varrho_left2, varrho_left_dict,  varrho_right1, varrho_right2, varrho_right_dict = main(CR,  w)
    print("D^{'}:")
    print("===========")
    print_dataset(D)
    print("Varrho Left:")
    print("===========")
    print([str(interval) for interval in varrho_left1])
    print([str(interval) for interval in varrho_left2])
    for key, value in varrho_left_dict.items():
        print(str(key), [str(item) for item in value])
    print("Varrho Right:")
    print("===========")
    print([str(interval) for interval in varrho_right1])
    print([str(interval) for interval in varrho_right2])

    for key, value in varrho_right_dict.items():
        print(str(key), [str(item) for item in value])











from collections import defaultdict
from meteor_reasoner.classes.interval import Interval
from meteor_reasoner.classes.atom import Atom
from meteor_reasoner.classes.literal import *
from meteor_reasoner.materialization.coalesce import coalescing


def trim_delta(window_facts, delta_new, limit):
    trimmed_delta_new = defaultdict(lambda: defaultdict(list))
    for predicate in delta_new:
        for entity in delta_new[predicate]:
            tmp_interval_list = []
            if predicate in window_facts and entity in window_facts[predicate]:
                diff_interval_list = Interval.diff_list(delta_new[predicate][entity], window_facts[predicate][entity])
                if len(diff_interval_list) > 0:
                    for interval in diff_interval_list:
                        new_interval = Interval.intersection(limit, interval)
                        if new_interval is not None:
                            tmp_interval_list.append(new_interval)
                    if len(tmp_interval_list) != 0:
                        trimmed_delta_new[predicate][entity] = tmp_interval_list
            else:
                for interval in delta_new[predicate][entity]:
                    new_interval = Interval.intersection(limit, interval)
                    if new_interval is not None:
                        tmp_interval_list.append(new_interval)
                if len(tmp_interval_list) != 0:
                    trimmed_delta_new[predicate][entity] = tmp_interval_list

    return trimmed_delta_new


def trim_window(window_facts, limit):
    trimmed_delta_new = defaultdict(lambda: defaultdict(list))
    for predicate in window_facts:
        for entity in window_facts[predicate]:
            tmp_interval_list = []
            for interval in window_facts[predicate][entity]:
                new_interval = Interval.intersection(limit, interval)
                if new_interval is not None:
                    tmp_interval_list.append(new_interval)
            if len(tmp_interval_list) != 0:
                trimmed_delta_new[predicate][entity] = tmp_interval_list
    return trimmed_delta_new


def contain_new(delta_new, window_facts, limit):
    for predicate in delta_new:
        for entity in delta_new[predicate]:
            if predicate in window_facts and entity in window_facts[predicate]:
                diff_interval_list = Interval.diff_list(delta_new[predicate][entity], window_facts[predicate][entity])
                for interval in diff_interval_list:
                    new_interval = Interval.intersection(limit, interval)
                    if new_interval is not None:
                        return True
            else:
                for interval in delta_new[predicate][entity]:
                    new_interval = Interval.intersection(limit, interval)
                    if new_interval is not None:
                        return True
    return False


def merge_streams(window_facts, delta):
    for predicate in delta:
        for entity in delta[predicate]:
            window_facts[predicate][entity] = coalescing(window_facts[predicate][entity] + delta[predicate][entity])


def add_streams(window_facts, streams, t_next):
    for stream in streams:
        window_facts[stream.predicate][stream.entity] = coalescing(window_facts[stream.predicate][stream.entity]+[Interval(t_next, t_next, False, False)])


def get_maximum_rational_number(program):
    # assume there is no nexted operators, and only boxminus, boxplus, diamondminus and
    # diamondplus are included and no bottom and top

    maximum_val = 0
    for rule in program:
        head = rule.head
        body = rule.body
        for literal in [head]+body:
            if isinstance(literal, Atom): # no operator
                continue
            elif isinstance(literal, Literal):
                if len(literal.operators) > 0:
                    maximum_val = max(maximum_val, literal.operators[0].interval.right_value)
    return maximum_val





from meteor_reasoner.materialization.join_util import *
from meteor_reasoner.materialization.apply import *


def ground_generator(literal, context, D, D_index=None):
    """
    This function is a generator, which generates entity and the context.
    Args:
        literal (a literal instance):
        context (a dictionary): store the mapping of  variables and constants
        D (a dictionary object): a global object storing all facts.

    Returns:
        entity and the context

    """
    predicate = literal.get_predicate()
    entity = copy.deepcopy(literal.get_entity())

    if predicate not in D:
        return
    if len(entity) == 1 and entity[0].name == "nan":
        yield entity, dict()

    elif not contain_variable(entity):
        if predicate in D and entity in D[predicate]:
            yield entity, dict()

    elif not contain_variable_after_replace(entity, context):
        replaced_entity = []
        for term in entity:
            if term.type == "variable":
                term.type = "constant"
                term.name = context[term.name]
                replaced_entity.append(term)
            else:
                replaced_entity.append(term)

        if predicate in D and tuple(replaced_entity) in D[predicate]:
            yield tuple(replaced_entity), dict()
    else:
        if D_index is not None:
            index_str = []
            for i, term in enumerate(entity):
                if term.type == "constant":
                    index_str.append(str(i) + "@" + term.name)
                else:
                    if term.name in context:
                        index_str.append(str(i)+"@"+context[term.name])

            index_str = "||".join(index_str)
            if len(index_str) == 0:
                for constant_entity in D[predicate]:
                    tmp_context = dict()
                    for term1, term2 in zip(entity, constant_entity):
                        if term1.type == "constant" and term1.name != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                            break
                        elif term1.type == "variable" and term1.name in tmp_context and tmp_context[term1.name] != term2.name:
                            break
                        else:
                            if term1.type == "variable":
                                tmp_context[term1.name] = term2.name
                    else:
                        yield constant_entity, tmp_context
            else:
                if index_str in D_index[predicate]:
                    for constant_entity in D_index[predicate][index_str]:
                        tmp_context = dict()
                        for term1, term2 in zip(entity, constant_entity):
                            if term1.type == "constant" and term1.name != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                                break
                            elif term1.type == "variable" and term1.name in tmp_context and tmp_context[term1.name] != term2.name:
                                break
                            else:
                                if term1.type == "variable":
                                    tmp_context[term1.name] = term2.name
                        else:
                            yield constant_entity, tmp_context

        else:
            for tmp_entity in D[predicate]:
                tmp_context = dict()
                for term1, term2 in zip(entity, tmp_entity):
                    if term1.type == "constant" and term1.name != term2.name:
                        break
                    elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                        break
                    elif term1.type == "variable" and term1.name in tmp_context and tmp_context[term1.name] != term2.name:
                        break
                    else:
                        if term1.type == "variable":
                            tmp_context[term1.name] = term2.name
                else:
                    yield tmp_entity, tmp_context


if __name__ == "__main__":
    from collections import defaultdict
    from meteor_reasoner.classes import Term
    from meteor_reasoner.materialization.index_build import build_index

    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike"), Term("nick")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]
    D_index = build_index(D)

    Delta = defaultdict(lambda: defaultdict(list))
    head = Atom("C", tuple([Term("nan")]))
    literal_a = Literal(Atom("A", tuple([Term("Y", "variable"), Term("X", "variable")])), [Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])

    body = [literal_a, literal_b]

    new_literal = BinaryLiteral(Atom("A", tuple([Term("X", "variable"), Term("Y", "variable")])), Atom("B", tuple([Term("nan")])), Operator("Since", Interval(1, 2, False, False)))
    body.append(new_literal)

    for atom_instance in ground_generator(literal_a, context={}, D=D, D_index=D_index):
        print(atom_instance)






from meteor_reasoner.classes.literal import *
from meteor_reasoner.classes.interval import *
from meteor_reasoner.classes.atom import *
import copy


def since_deduce(literal, left_interval, right_interval):
    """
    Calculate the ``Since'' operator .
    Args:
        literal (a BinarayLiteral instance):
        left_interval (an Interval instance):
        right_interval (an Interval instance):
    Returns:
        a Interval instance or None
    """
    if not literal.operator.interval.left_open and literal.operator.interval.left_value == 0:
        return Interval.add(right_interval, literal.operator.interval)

    # calculate S
    closed_left_interval = copy.copy(left_interval)
    closed_left_interval.left_open = False
    closed_left_interval.right_open = False

    s_interval = Interval.intersection(closed_left_interval, right_interval)
    if s_interval is None:
        return None

    # calculate T satisfying the right literal
    t_interval = Interval.add(s_interval, literal.operator.interval)

    # calculate T satisfying the left literal
    interval = Interval.intersection(t_interval, closed_left_interval)
    if interval is None:
        return None
    else:
        return interval


def until_deduce(literal, left_interval, right_interval):
    """
    Calculate the ``Until'' operator .
    Args:
        literal (a BinarayLiteral instance):
        left_interval (an Interval instance):
        right_interval (an Interval instance):

    Returns:
        a Interval instance or None
    """
    if not literal.operator.interval.left_open and literal.operator.interval.left_value == 0:
        return Interval.sub(right_interval, literal.operator.interval)

    # calculate S
    closed_left_interval = copy.copy(left_interval)
    closed_left_interval.left_open = False
    closed_left_interval.right_open = False

    s_interval = Interval.intersection(closed_left_interval, right_interval)
    if s_interval is None:
        return None

    # calculate T satisfying the right literal
    t_interval = Interval.sub(s_interval, literal.operator.interval)

    # calculate T satisfying the left literal
    interval = Interval.intersection(t_interval,closed_left_interval)
    if interval is None:
        return None
    else:
        return interval


def apply(literal, D):
    """
    Apply MTL operator(s) to a literal.
    Args:
        literal (Literal BinaryLiteral Instance):
        D (defaultdict):

    Returns:
        A list of Interval instances.
    """
    if not isinstance(literal, BinaryLiteral) and (isinstance(literal, Atom) or
                                                   len(literal.operators) == 0):
        predicate = literal.get_predicate()
        entity = literal.get_entity()

        if entity: # not None
            if predicate in D and entity in D[predicate]:
                return D[predicate][entity]
            else:
                return []

        else:
            if predicate in D:
                return D[predicate]
            else:
                return []
    else:
        op_name = literal.get_op_name()
        if op_name in ["Until", "Since"]:
            left_literal = literal.left_literal
            right_literal = literal.right_literal

            if left_literal.get_predicate() == "Bottom" or right_literal.get_predicate() == "Bottom":
                raise ValueError("It is not allowed to have the bottom predicate in the left side "
                                 "or the right side of a binary literal ")
            if left_literal.get_predicate() == "Top":
                T1 = [Interval(float('-inf'), float('inf'), True, True)]
            else:
                T1 = apply(left_literal, D)

            if right_literal.get_predicate() == "Top":
                T2 = [Interval(float('-inf'), float('inf'), True, True)]
            else:
                T2 = apply(right_literal, D)

            T = []
            if op_name == "Until":
                if T1 and T2:
                    for t1 in T1:
                        for t2 in T2:
                            t = until_deduce(literal, t1, t2)
                            if t:
                                T.append(t)
            else:
                if T1 and T2:
                    for t1 in T1:
                        for t2 in T2:
                            t = since_deduce(literal, t1, t2)
                            if t:
                                T.append(t)
            return T

        else:
            pop_operator = literal.operators.pop(0)
            T0 = apply(literal, D)
            literal.operators.insert(0, pop_operator)
            T = []
            if op_name == "Diamondminus":
                for t0 in T0:
                  interval = Interval.add(t0, literal.operators[0].interval)
                  if Interval.is_valid_interval(interval.left_value, interval.right_value,
                                                interval.left_open, interval.right_open):
                      T.append(interval)
            elif op_name == "Boxminus":
                for t0 in T0:
                  interval = Interval.circle_add(t0, literal.operators[0].interval)
                  if Interval.is_valid_interval(interval.left_value, interval.right_value,
                                                interval.left_open, interval.right_open):
                      T.append(interval)
            elif op_name == "Diamondplus":
                for t0 in T0:
                  interval = Interval.sub(t0, literal.operators[0].interval)
                  if Interval.is_valid_interval(interval.left_value, interval.right_value,
                                                interval.left_open, interval.right_open):
                      T.append(interval)
            elif op_name == "Boxplus":
                for t0 in T0:
                  interval = Interval.circle_sub(t0, literal.operators[0].interval)
                  if Interval.is_valid_interval(interval.left_value, interval.right_value,
                                                interval.left_open, interval.right_open):
                      T.append(interval)
            else:
                raise ValueError("{} is an illegal MTL operator name!".format(op_name))

            return T


def reverse_apply(literal, D):
    """
    Apply MTL operator(s) to a literal.

    Args:
        literal (Literal BinaryLiteral Instance):
        D (defaultdict):

    Returns:
        A list of Interval instances.
    """
    if isinstance(literal, Atom) or len(literal.operators) == 0:
        predicate = literal.get_predicate()
        entity = literal.get_entity()

        if entity: # not None
            if predicate in D and entity in D[predicate]:
                return D[predicate][entity]
            else:
                return []

        else:
            if predicate in D:
                return D[predicate]
            else:
                return []
    else:
        op_name = literal.get_op_name()
        pop_operator = literal.operators.pop(0)
        T0 = reverse_apply(literal, D)
        literal.operators.insert(0, pop_operator)
        T = []
        if op_name == "Boxplus":
            for t0 in T0:
              interval = Interval.add(t0, literal.operators[0].interval)
              if Interval.is_valid_interval(interval.left_value, interval.right_value,
                                            interval.left_open, interval.right_open):
                  T.append(interval)

        elif op_name == "Boxminus":
            for t0 in T0:
              interval = Interval.sub(t0, literal.operators[0].interval)
              if Interval.is_valid_interval(interval.left_value, interval.right_value,
                                            interval.left_open, interval.right_open):
                  T.append(interval)
        else:
            raise ValueError("{} is an illegal MTL operator name!".format(op_name))

        return T


if __name__ == "__main__":
    from collections import defaultdict
    from meteor_reasoner.classes import Term
    from meteor_reasoner.materialization.index_build import build_index

    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike"), Term("nick")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]
    D_index = build_index(D)
    literal_a = Literal(Atom("A", tuple([Term("mike"), Term("nick")])), [Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("A", tuple([Term("mike"), Term("nick")])), [Operator("Diamondminus", Interval(1, 2, False, False))])
    bi_literal_a = BinaryLiteral(Atom("A", tuple([Term("mike"), Term("nick")])),
                               Atom("B", tuple([Term("nan")])), Operator("Since", Interval(1, 2, False, False)))
    bi_literal_b = BinaryLiteral(literal_a, literal_b, Operator("Since", Interval(1, 2, False, False)))

    t1 = apply(literal_a, D)
    for t in t1:
        print(t)
    print("end")
    t2 = reverse_apply(literal_a, D)
    for t in t2:
        print(t)
    print("end")
    t3 = apply(bi_literal_a, D)
    for t in t3:
        print(t)
    print("end")
    t4 = apply(bi_literal_b, D)
    for t in t4:
        print(t)
    print("end")






class Term:
    def __init__(self, name, type="constant"):
        """
        Args:
            value (string):
            type (str): ``constant '' or ``variable''
        """
        self.name = name
        self.type = type

    @classmethod
    def new_term(cls, name, type="constant"):
        return cls(name, type)

    def __hash__(self):
        return hash(self.type + str(self.name))

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, Term):
            return self.type == other.type and self.name == other.name

        return False



from meteor_reasoner.utils.parser import *


def test_literal():
    raw_literal = "Boxminus[1,2]ASince[1,2]Diamondminus[3,4]Boxminus[5,6]B(X)"
    literal = parse_literal(raw_literal)
    assert str(literal) == raw_literal

    raw_literal = "ASince[1,2]Diamondminus[3,4]Boxminus[5,6]B(X)"
    literal = parse_literal(raw_literal)
    assert str(literal) == raw_literal


def test_rule():
    raw_rule = "Boxplus[1,3]Boxminus[1,2]A:-Diamondminus[1,2]B(X),Boxminus[1,2]Diamondminus[2,3]C(X)"
    rule = parse_rule(raw_rule)
    print(str(rule))
    assert raw_rule == str(rule)




from meteor_reasoner.classes.atom import Atom
from meteor_reasoner.classes.literal import Literal, BinaryLiteral


def check_propagation(program):
    forward = True
    backward = True
    for rule in program:
        head = rule.head
        if isinstance(head, Literal):
            for operator in head.operators:
                if operator.name in ["Boxplus"]:
                    backward = False
                elif operator.name in ["Boxminus"]:
                    forward = False
        if not forward and not backward:
            return 0 # neither forwarding or backwarding

        for literal in rule.body:
            if isinstance(literal, Literal):
                for operator in literal.operators:
                    if operator.name in ["Boxminus", "Diamondminus"]:
                        backward = False
                    elif operator.name in ["Boxplus", "Diamondminus"]:
                        forward = False
            elif isinstance(literal, BinaryLiteral):
                if literal.operator.name in ["Boxminus", "Diamondminus"]:
                    backward = False
                elif literal.operator.name in ["Boxplus", "Diamondminus"]:
                    forward = False
                left_literal = literal.left_literal
                if not isinstance(left_literal, Atom):
                    for operator in left_literal.operators:
                        if operator.name in ["Boxminus", "Diamondminus"]:
                            backward = False
                        elif operator.name in ["Boxplus", "Diamondminus"]:
                            forward = False
                right_literal = literal.right_literal
                if not isinstance(right_literal, Atom):
                    for operator in right_literal.operators:
                        if operator.name in ["Boxminus", "Diamondminus"]:
                            backward = False
                        elif operator.name in ["Boxplus", "Diamondminus"]:
                            forward = False

        if not forward and not backward:
            return 0  # neither forwarding or backwarding
    if forward:
        return 1 # forwarding
    elif backward:
        return 2 # backwarding
    else:
        return 0  # mix









from meteor_reasoner.materialization.t_operator import *
from meteor_reasoner.materialization.coalesce import *
from meteor_reasoner.materialization.index_build import *
from meteor_reasoner.utils.propagation import check_propagation
from meteor_reasoner.graphutil.temporal_dependency_graph import CycleFinder
import time
from meteor_reasoner.materialization.utils import no_new_facts
from meteor_reasoner.utils.operate_dataset import save_dataset_to_file

def calculate_redundancy(delta, old):
    cnt = 0
    for predicate in delta:
        for entity in delta[predicate]:
            if predicate not in old or entity not in old[predicate]:  # all fact w.r.t the predicate are redundant
                cnt += len(delta[predicate][entity])
            else:
                for delta_interval in delta[predicate][entity]:
                    for old_interval in old[predicate][entity]:
                        if Interval.intersection(delta_interval, old_interval) is not None:
                            break
                    else:
                        cnt += 1
    return cnt


def seminaive_combine(D, delta_new, delta_old, D_index=None):
    for head_predicate in delta_new:
        for entity, T in delta_new[head_predicate].items():
            if head_predicate not in D or entity not in D[head_predicate]:
                D[head_predicate][entity] = D[head_predicate][entity] + T
                delta_old[head_predicate][entity] = delta_old[head_predicate][entity] + T
                # update index
                for i, item in enumerate(entity):
                    D_index[head_predicate][str(i) + "@" + item.name].append(entity)
                if len(entity) > 2:
                    for i, item1 in enumerate(entity):
                        for j, item2 in enumerate(entity):
                            if j <= i:
                                continue
                            D_index[head_predicate][str(i) + "@" + item1.name + "||" + str(j) + "@" + item2.name].append(entity)
            else:
                coalesced_T = coalescing(T + D[head_predicate][entity])
                if not Interval.compare(coalesced_T, D[head_predicate][entity]):
                    for interval1 in coalesced_T:
                        flag = True
                        for interval2 in D[head_predicate][entity]:
                            if interval1.left_value == interval2.left_value and interval1 == interval2:
                                flag = False
                                break
                            elif interval2.left_value > interval1.left_value:
                                break
                        if flag:
                            delta_old[head_predicate][entity].append(interval1)
                    D[head_predicate][entity] = coalesced_T
    fixpoint = True
    if len(delta_old) != 0:
        coalescing_d(D)
        coalescing_d(delta_old)
        fixpoint = False
    return fixpoint


def naive_combine(D, delta_new, D_index=None):
    fixpoint = True
    for head_predicate in delta_new:
        for entity, T in delta_new[head_predicate].items():
            if head_predicate not in D or entity not in D[head_predicate]:
                fixpoint = False
                try:
                    D[head_predicate][entity] = D[head_predicate][entity] + T
                except:
                    print("debug")
                    D[head_predicate][entity] = D[head_predicate][entity] + T
                # update index
                for i, item in enumerate(entity):
                    D_index[head_predicate][str(i) + "@" + item.name].append(entity)
                if len(entity) > 2:
                    for i, item1 in enumerate(entity):
                        for j, item2 in enumerate(entity):
                            if j <= i:
                                continue
                            D_index[head_predicate][str(i) + "@" + item1.name + "||" + str(j) + "@" + item2.name].append(entity)
            else:
                coalesced_T = coalescing(T + D[head_predicate][entity])
                if fixpoint:
                    if coalesced_T != D[head_predicate][entity]:
                        fixpoint = False
                D[head_predicate][entity] = coalesced_T

    if not fixpoint:
         coalescing_d(D)

    return fixpoint


def materialize(D, rules, D_index, seminaive=False, delta_old=None, K=100, logger=None, recorder=None):
    """
    The function implements the materialization operation.
    Args:
        D (dictionary object):
        rules (list of Rule instances):
        K (int): the nunber of steps for exectuting the materialization. By default=100.
    Returns:
        The
    """
    k = 0
    if logger is not None:
        start_time = time.time()
        calc_time = 0.0

    while k < K:
        k += 1
        if seminaive:
            semi_start_time = time.time()
            delta_new = seminaive_immediate_consequence_operator(rules, D, D_index, k=k, delta_old=delta_old)
            #print("One semi round:", time.time()-semi_start_time)
        else:
            naive_start_time = time.time()
            delta_new = naive_immediate_consequence_operator(rules, D, D_index, recorder=recorder)
            #print("One naive round:", time.time() - naive_start_time)

        if seminaive:
            delta_old = defaultdict(lambda: defaultdict(list))
            semi_combine_start_time = time.time()
            fixpoint = seminaive_combine(D, delta_new, delta_old, D_index)
            if logger is not None:
                tmp_start_time = time.time()
                coalescing_d(delta_new)
                number_of_redundant_facts = calculate_redundancy(delta_new, delta_old)
                total_number = 0
                for predicate in D:
                    for entity in D[predicate]:
                        total_number += len(D[predicate][entity])
                calc_time += time.time() - tmp_start_time
                logger.info("Iteration={}, t={}, D={}, n={}".format(k, time.time() - start_time - calc_time, total_number, number_of_redundant_facts))
        else:
            if logger is not None:
                delta_old = defaultdict(lambda: defaultdict(list))
                naive_combine_start_time = time.time()
                fixpoint = seminaive_combine(D, delta_new, delta_old, D_index)
                tmp_start_time = time.time()
                coalescing_d(delta_new)
                number_of_redundant_facts = calculate_redundancy(delta_new, delta_old)
                total_number = 0
                for predicate in D:
                    for entity in D[predicate]:
                        total_number += len(D[predicate][entity])
                calc_time += time.time() - tmp_start_time
                logger.info("Iteration={}, t={}, D={}, n={}".format(k, time.time() - start_time - calc_time, total_number, number_of_redundant_facts))
            else:
                fixpoint = naive_combine(D, delta_new, D_index)

        if fixpoint:
            save_dataset_to_file("recursive_dataset.txt", D)
            return True

    return False


def opt_materialize(D, rules, D_index, min_value=decimal.Decimal(0), max_value=decimal.Decimal(10000), delta_old=None, K=100, logger=None):
    # preprocessing
    CF = CycleFinder(program=rules)
    non_predicates = CF.get_non_recursive_predicates()
    #print("Non-recursive predicates:", non_predicates)
    nr_program = list()
    r_program = list()
    for rule in rules:
        if rule.head.get_predicate() in non_predicates:
            nr_program.append(rule)
        else:
            r_program.append(rule)
    # for rule in nr_program:
    #     print("rule:", rule)
    # exit()
    while True:
        delta_new = seminaive_immediate_consequence_operator(nr_program, D, D_index, delta_old=delta_old)
        delta_old = defaultdict(lambda: defaultdict(list))
        fixpoint = seminaive_combine(D, delta_new, delta_old, D_index)
        if fixpoint:
            break
    propagation = check_propagation(nr_program)
    # forward propagation
    observed_rules = defaultdict(list)
    if propagation == 1:
        print("Is a forward-propagating program!")
        for rule in r_program:
            min_right = float("inf")
            for literal in rule.body:
                if isinstance(literal, BinaryLiteral):
                    predicate = literal.left_literal.get_predicate()
                    if predicate in non_predicates:
                        tmp_max_right = float("-inf")
                        for entity in D[predicate]:
                            tmp_max_right = max(tmp_max_right, max([interval.right_value for interval in D[predicate][entity]]))

                        min_right = min(min_right, tmp_max_right)
                    predicate = literal.right_literal.get_predicate()
                    if predicate in non_predicates:
                        tmp_max_right = float("-inf")
                        for entity in D[predicate]:
                            tmp_max_right = max(tmp_max_right,
                                                max([interval.right_value for interval in D[predicate][entity]]))
                        min_right = min(min_right, tmp_max_right)
                else:
                    predicate = literal.get_predicate()
                    if predicate in non_predicates:
                        tmp_max_right = float("-inf")
                        for entity in D[predicate]:
                            tmp_max_right = max(tmp_max_right,
                                                max([interval.right_value for interval in D[predicate][entity]]))
                        min_right = min(min_right, tmp_max_right)

            if min_right != float("inf"):
                observed_rules[min_right].append(rule)

    # backward propagation
    elif propagation == 2:
        print("Is a backward-propagating program!")
        for rule in r_program:
            max_left = float("-inf")
            for literal in rule.body:
                if isinstance(literal, BinaryLiteral):
                    predicate = literal.left_literal.get_predicate()
                    if predicate in non_predicates:
                        tmp_min_left = float("inf")
                        for entity in D[non_predicates]:
                           tmp_min_left = min(tmp_min_left, min([interval.left_value for interval in D[predicate][entity]]))

                        max_left = max(max_left, tmp_min_left)
                    predicate = literal.right_literal.get_predicate()
                    if predicate in non_predicates:
                        tmp_min_left = float("inf")
                        for entity in D[non_predicates]:
                            tmp_min_left = min(tmp_min_left,
                                               min([interval.left_value for interval in D[predicate][entity]]))

                        max_left = max(max_left, tmp_min_left)
                else:
                    predicate = literal.get_predicate()
                    if predicate in non_predicates:
                        tmp_min_left = float("inf")
                        for entity in D[non_predicates]:
                            tmp_min_left = min(tmp_min_left,
                                               min([interval.left_value for interval in D[predicate][entity]]))

                        max_left = max(max_left, tmp_min_left)

            if max_left is not float("-inf"):
                observed_rules[rule] = max_left

    observed_rules = sorted(observed_rules.items(), key=lambda item: item[0])
    # for key, value in observed_rules:
    #     print(key, [str(v) for v in value])
    # print("=====================")
    # exit()
    k=0
    records = {}
    if logger is not None:
       start_time = time.time()
       calc_time = 0.0

    delta_old = D
    while k < K:
        k += 1
        if k == 9:
            r_program = r_program[-8:]
        delta_new = seminaive_immediate_consequence_operator(r_program, D, D_index, k=k,
                                                                      delta_old=delta_old)

        if propagation == 1:
            for i, item in enumerate(observed_rules):
                if no_new_facts(delta_new, D, Interval(min_value, item[0], False, False)):
                    for rule in item[1]:
                        #print("Removed rule:", str(rule))
                        try:
                          r_program.remove(rule)
                        except:
                            continue
                        #print("Number of remaining rules:", len(r_program))

                else:
                    observed_rules = observed_rules[i:]
                    break
            else:
                observed_rules = observed_rules[i+1:]

        elif propagation == 2:
            for i, item in enumerate(observed_rules):
                if no_new_facts(delta_new, D, Interval(item[0], max_value, False, False)):
                    for rule in item[1]:
                        r_program.remove(rule)
                else:
                    observed_rules = observed_rules[i:]
                    break
            else:
                observed_rules = observed_rules[i + 1:]

        delta_old = defaultdict(lambda: defaultdict(list))
        fixpoint = seminaive_combine(D, delta_new, delta_old, D_index)

        if logger is not None:
            tmp_start_time = time.time()
            coalescing_d(delta_new)
            number_of_redundant_facts = calculate_redundancy(delta_new, delta_old)
            total_number = 0
            for predicate in D:
                for entity in D[predicate]:
                    total_number += len(D[predicate][entity])
            calc_time += time.time()-tmp_start_time
            logger.info("Iteration={}, t={}, D={}, n={}".format(k, time.time() - start_time-calc_time, total_number, number_of_redundant_facts))
        if fixpoint:
            return True
    return False


if __name__ == "__main__":
    from meteor_reasoner.utils.operate_dataset import print_dataset
    from meteor_reasoner.utils.loader import *

    start_time = time.time()
    raw_program = [
        "P(X,Y):-Diamondminus[0,1]P(X,Y)",
        "R(Y):-Q(Y),H(Z),P(X,Y)",
        "Q(X):-K(X,Y)"
    ]
    raw_data = [
        "P(a,b)@0",
        "K(b,c)@[1,2]",
        "H(c)@[2,5]"
    ]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    program = load_program(raw_program)
    opt_materialize(D, rules=program, D_index=D_index, delta_old=D, K=1000)
    print("Optimized method time: ", time.time()-start_time)

    start_time = time.time()
    raw_program = [
        "P(X,Y):-Diamondminus[0,1]P(X,Y)",
        "R(Y):-Q(Y),H(Z),P(X,Y)",
        "Q(X):-K(X,Y)"
    ]
    raw_data = [
        "P(a,b)@0",
        "K(b,c)@[1,2]",
        "H(c)@[2,5]"
    ]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    program = load_program(raw_program)
    materialize(D, rules=program, D_index=D_index, delta_old=D, seminaive=True, K=1000)
    print("Seminaive method time: ", time.time() - start_time)

    start_time = time.time()
    raw_program = [
        "P(X,Y):-Diamondminus[0,1]P(X,Y)",
        "R(Y):-Q(Y),H(Z),P(X,Y)",
        "Q(X):-K(X,Y)"
    ]
    raw_data = [
        "P(a,b)@0",
        "K(b,c)@[1,2]",
        "H(c)@[2,5]"
    ]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    program = load_program(raw_program)
    materialize(D, rules=program, D_index=D_index, K=1000)
    print("naive method time: ", time.time() - start_time)


    exit()
    raw_data = ["P(mike,mike)@1"]
    raw_program = ["P(X,X):- Diamondminus[1,1]P(X,X)"]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    Program = load_program(raw_program)

    start_time = time.time()
    materialize(D, Program, D_index, K=100)
    print_dataset(D)
    end_time = time.time()
    print("The naive mat time:", end_time-start_time)

    raw_data = ["P(mike,mike)@1"]
    raw_program = ["P(X,X):- Diamondminus[1,1]P(X,X)"]
    D = load_dataset(raw_data)
    coalescing_d(D)
    D_index = build_index(D)
    Program = load_program(raw_program)
    start_time = time.time()
    materialize(D, Program, D_index, seminaive=True, delta_old=D, K=100)
    print_dataset(D)
    end_time = time.time()
    print("The seminaive mat time:", end_time - start_time)



















from meteor_reasoner.materialization.seminaive_join import *
from meteor_reasoner.materialization.naive_join import *
from meteor_reasoner.materialization.optimized_seminaive_join_bak import *
from collections import defaultdict
import time


def naive_immediate_consequence_operator(rules, D, D_index, recorder=None):
    delta_new = defaultdict(lambda: defaultdict(list))
    for i, rule in enumerate(rules):
        #print("Rule application:", str(rule))
        t_start = time.time()
        naive_join(rule, D, delta_new, D_index, recorder=recorder)
        #print("Run time:", time.time()-t_start)
    return delta_new


def seminaive_immediate_consequence_operator(rules, D, D_index, k=1, delta_old=None):
    delta_new = defaultdict(lambda: defaultdict(list))
    for i, rule in enumerate(rules):
        #print("Rule application:", str(rule))
        t_start = time.time()
        seminaive_join(rule, D, delta_old, delta_new, k=k, D_index=D_index)
        #print("Run time:", time.time() - t_start)

    return delta_new


def optimize_seminaive_immediate_consequence_operator(rules, D, D_index, delta_old=None, non_predicates=[], records={}):
    delta_new = defaultdict(lambda: defaultdict(list))
    for i, rule in enumerate(rules):
        optimized_seminaive_join(rule, D, delta_old, delta_new, D_index, records=records, non_predicates=non_predicates)

    return delta_new


import networkx as nx


class TemporalDependencyGraph():
    def __init__(self):
        self.G = nx.MultiDiGraph()

    def add_edge(self, node1, node2, value):
        self.G.add_edge(node1, node2, weight=value)

    def add_edge_list(self, edges):
        for node1, node2, value in edges:
            self.G.add_edge(node1, node2, weight=2)

    @staticmethod
    def allzerocycle(results):
        def helper(i, path):
            if i == len(results):
                if sum(path) != 0:
                    return -1
                else:
                    return 0

            for item in results[i]:
                res = helper(i+1, path+[item])
                if res == -1:
                    return -1

        res = helper(0, [])
        if res is not None:
            return True
        else:
            return False

    def is_cyclic(self):
        visited = set()
        for node in self.G:
            # judge whether there exist a self-loop
            node_edge_data = self.G.get_edge_data(node, node)
            if node_edge_data is not None:
                for value in node_edge_data.values():
                    if value != 0:
                        return True

            descendants = nx.algorithms.descendants(self.G, node)
            for child in descendants:
                if node in self.G.neighbors(child):  #exist a loop
                    all_simple_paths = set([tuple(path[1:] + [node]) for path in nx.all_simple_paths(self.G, node, child)])
                    all_simple_paths = [path for path in all_simple_paths if tuple(sorted(path)) not in visited]
                    for path in all_simple_paths:
                        visited.add(tuple(sorted(path)))
                        last_node = node
                        path_weights_list = []
                        for path_node in path:
                            weight_tuple = []
                            for value in self.G.get_edge_data(last_node, path_node).values():
                                if value["weight"][0] != value["weight"][1]:
                                    return True
                                weight_tuple.append(value["weight"][0])
                            path_weights_list.append(weight_tuple)
                            last_node = path_node

                        result = self.allzerocycle(path_weights_list)
                        if result:
                            return True
        return False


from decimal import Decimal
from meteor_reasoner.classes.interval import Interval


class CommonFragment:
    def __init__(self, base_interval, cr_flag=True):
        """
        Args:
            base_interval: [t_D^-, t_D^+]
            common:
            cr_flag:
        """
        self.base_interval = base_interval
        self.common = None
        self.cr_flag = cr_flag


from meteor_reasoner.classes.interval import *


class AutomataWindow:
    def __init__(self, ruler_intervals, ruler_intervals_literals, left_pattern, right_pattern):
        """
        Construct a Window consists of a list of Interval instances and Literal and Atom instances at each Interval instance.
        Args:
            ruler_intervals (list of Interval instances):
            ruler_intervals_literals (dictionary object): the key is the Interval and the value is the set of Literal and Atom instances
            left_pattern:
            right_pattern:
        """
        self.left_pattern = left_pattern
        self.right_pattern = right_pattern
        self.ruler_intervals = ruler_intervals
        self.ruler_intervals_literals = ruler_intervals_literals

    def __eq__(self, other):
        if isinstance(other, AutomataWindow):
            self_interval_len = [str(item.right_value - item.left_value) for item in self.ruler_intervals]
            other_interval_len = [str(item.right_value - item.left_value) for item in other.ruler_intervals]

            if self_interval_len != other_interval_len:
                return False

            self_interval_literal_str = []
            for i, ruler_interval in enumerate(self.ruler_intervals):
                tmp = []
                for literal in self.ruler_intervals_literals[ruler_interval]:
                    tmp.append(str(i) + str(literal))
                tmp.sort()
                self_interval_literal_str.append("#".join(tmp))

            other_interval_literal_str = []
            for i, ruler_interval in enumerate(other.ruler_intervals):
                tmp = []
                for literal in other.ruler_intervals_literals[ruler_interval]:
                    tmp.append(str(i) + str(literal))
                tmp.sort()
                other_interval_literal_str.append("#".join(tmp))

            if "$".join(self_interval_literal_str) == "$".join(other_interval_literal_str):
                return True
            else:
                return False

        return False

    def __str__(self):
       interval_len = [str(item.right_value - item.left_value) for item in self.ruler_intervals]
       interval_literal_str = []
       for i, ruler_interval in enumerate(self.ruler_intervals):
          tmp = []
          for literal in self.ruler_intervals_literals[ruler_interval]:
             tmp.append(str(i) + str(literal))
          tmp.sort()
          interval_literal_str.append("#".join(tmp))
       return "#".join(interval_len) + "$".join(interval_literal_str)

    def __hash__(self):
       interval_len = [str(item.right_value - item.left_value) for item in self.ruler_intervals]
       interval_literal_str = []
       for i, ruler_interval in enumerate(self.ruler_intervals):
          tmp = []
          for literal in self.ruler_intervals_literals[ruler_interval]:
             tmp.append(str(i) + str(literal))
          tmp.sort()
          interval_literal_str.append("#".join(tmp))
       return hash("#".join(interval_len) + "$".join(interval_literal_str))


    def remove_the_most_right_ruler_interval(self):
       the_most_right_ruler_interval = self.ruler_intervals[-1]
       del self.ruler_intervals_literals[the_most_right_ruler_interval]
       del self.ruler_intervals[-1]


    def remove_the_most_left_ruler_interval(self):
       the_most_left_ruler_interval = self.ruler_intervals[0]
       del self.ruler_intervals_literals[the_most_left_ruler_interval]
       del self.ruler_intervals[0]

    def move_left(self, subset_literal=[]):
       try:
          the_most_left_ruler_interval = self.ruler_intervals[0]
       except:
           print("debug")
       if the_most_left_ruler_interval.left_open:
          new_ruler_interval = Interval(the_most_left_ruler_interval.left_value, the_most_left_ruler_interval.left_value, False, False)
          self.ruler_intervals.insert(0, new_ruler_interval)
          self.ruler_intervals_literals[new_ruler_interval] = subset_literal
          self.remove_the_most_right_ruler_interval()

       else:
         # count the non-punctual ruler interval
         cnt = 0
         ruler_interval_len = []
         for i in range(len(self.ruler_intervals)):
            if self.ruler_intervals[i].left_open:
               ruler_interval_len.append(str(abs(self.ruler_intervals[i].right_value - self.ruler_intervals[i].left_value)))
               cnt += 1
            if cnt >= len(self.left_pattern):
               break

         while True:
            tmp_pattern = "#".join(ruler_interval_len)
            if tmp_pattern in self.left_pattern:
               next_ruler_interval_len = self.left_pattern[tmp_pattern]
               current_the_most_left_ruler_interval = self.ruler_intervals[0]
               new_ruler_interval = (Interval(current_the_most_left_ruler_interval.left_value - next_ruler_interval_len, current_the_most_left_ruler_interval.right_value, True, True))
               self.ruler_intervals.insert(0, new_ruler_interval)
               self.ruler_intervals_literals[new_ruler_interval] = subset_literal
               self.remove_the_most_right_ruler_interval()
               break
            del ruler_interval_len[-1]

    def move_right(self, subset_literal=[]):
      the_most_right_ruler_interval = self.ruler_intervals[-1]
      if the_most_right_ruler_interval.right_open:
         new_ruler_interval = Interval(the_most_right_ruler_interval.right_value, the_most_right_ruler_interval.right_value, False, False)
         self.ruler_intervals.append(new_ruler_interval)
         self.ruler_intervals_literals[new_ruler_interval] = subset_literal
         self.remove_the_most_left_ruler_interval()

      else:
         cnt = 0
         ruler_interval_len = []
         for i in range(len(self.ruler_intervals)):
            if self.ruler_intervals[i].left_open:
               ruler_interval_len.append(str(abs(self.ruler_intervals[i].right_value - self.ruler_intervals[i].left_value)))
               cnt += 1
            if cnt >= len(self.right_pattern):
               break

         while True:
            tmp_pattern = "#".join(ruler_interval_len)
            if tmp_pattern in self.right_pattern:
               next_ruler_interval_len = self.right_pattern[tmp_pattern]
               current_the_most_right_ruler_interval = self.ruler_intervals[-1]
               new_ruler_interval = Interval(current_the_most_right_ruler_interval.left_value, current_the_most_right_ruler_interval.left_value + next_ruler_interval_len, True, True)
               self.ruler_intervals.append(new_ruler_interval)
               self.ruler_intervals_literals[new_ruler_interval] = subset_literal
               self.remove_the_most_left_ruler_interval()
               break

            del ruler_interval_len[0]



from meteor_reasoner.classes.rule import *
from meteor_reasoner.materialization.ground import *
from meteor_reasoner.classes.term import Term
from collections import defaultdict
from meteor_reasoner.materialization.index_build import build_index
from meteor_reasoner.utils.operate_dataset import print_dataset
from meteor_reasoner.materialization.coalesce import coalescing_d


def seminaive_join(rule, D,  delta_old, delta_new,  k=1, D_index=None):
    """
    This function implement the join operator when variables exist in the body of the rule.
    Args:
        rule (a Rule instance):
        Delta (a dictionary object): store new materialized results.
        D (a dictionary object): a global variable storing all facts.

    Returns:
        None
    """
    head_entity = rule.head.get_entity()
    head_predicate = rule.head.get_predicate()
    literals = rule.body + rule.negative_body
    history_visited = defaultdict(int)

    def ground_body(global_literal_index, visited, delta, context):
        if global_literal_index == len(literals):
            for t_key, t_value in delta.items():
                t_item = str(t_key) + ":"
                for t_entity in t_value:
                    t_item += "-".join([str(item) for item in t_entity])
                history_visited[t_item] += 1

            T = []
            for i in range(len(rule.body)):
                grounded_literal = copy.deepcopy(literals[i])
                if isinstance(grounded_literal, BinaryLiteral):
                    grounded_literal.set_entity(delta[i])
                else:
                    if grounded_literal.get_predicate() not in ["Bottom", "Top"]:
                        grounded_literal.set_entity(delta[i][0])
                if i == visited:
                   t = apply(grounded_literal, D)
                else:
                   t = apply(grounded_literal, D)
                if len(t) == 0:
                    break
                else:
                    T.append(t)
            n_T = []
            for i in range(len(rule.body), len(literals)):
                grounded_literal = copy.deepcopy(literals[i])
                if isinstance(grounded_literal, BinaryLiteral):
                    grounded_literal.set_entity(delta[i])
                else:
                    if grounded_literal.get_predicate() not in ["Bottom", "Top"]:
                        grounded_literal.set_entity(delta[i][0])
                t = apply(grounded_literal, D)
                if len(t) == 0:
                    break
                else:
                    n_T.append(t)

            if len(n_T) > 0 and len(n_T) == len(rule.negative_body_atoms):
                n_T = interval_merge(T)
            else:
                n_T = []

            if head_entity is not None:
                replaced_head_entity = []
                for term in head_entity:
                    if term.type == "constant":
                        replaced_head_entity.append(term)
                    else:
                        if term.name not in context:
                            raise ValueError("Head variables in Rule({}) do not appear in the body".format(str(rule)))
                        else:
                            new_term = Term.new_term(context[term.name])
                            replaced_head_entity.append(new_term)
                replaced_head_entity = tuple(replaced_head_entity)
            else:
                replaced_head_entity = head_entity

            if len(T) == len(literals):
                T = interval_merge(T)
                exclude_t = []
                if len(T) != 0 and len(n_T) != 0:
                    exclude_t = interval_merge([T, n_T])
                if len(exclude_t) != 0:
                    T = Interval.diff(T, exclude_t)

                if len(T) != 0:
                    if not isinstance(rule.head, Atom):
                        tmp_D = defaultdict(lambda: defaultdict(list))
                        tmp_D[head_predicate][replaced_head_entity] = T
                        tmp_head = copy.deepcopy(rule.head)
                        tmp_head.set_entity(replaced_head_entity)
                        T = reverse_apply(tmp_head, tmp_D)

                    delta_new[head_predicate][replaced_head_entity] = delta_new[head_predicate][
                                                                          replaced_head_entity] + T

        elif global_literal_index == visited:
            ground_body(global_literal_index+1, visited, delta, context)

        else:
            current_literal = copy.deepcopy(literals[global_literal_index])
            if not isinstance(current_literal, BinaryLiteral):
                if current_literal.get_predicate() in ["Bottom", "Top"]:
                    ground_body(global_literal_index+1, delta, context)
                else:
                    for tmp_entity, tmp_context in ground_generator(current_literal, context, D, D_index):
                        tmp_delata = {global_literal_index: [tmp_entity]}
                        ground_body(global_literal_index+1,  visited, {**delta, **tmp_delata}, {**context, **tmp_context})

            else:
                left_predicate = current_literal.left_literal.get_predicate()
                right_predicate = current_literal.right_literal.get_predicate()

                if left_predicate in ["Bottom", "Top"]:
                    for tmp_entity, tmp_context in ground_generator(current_literal.right_literal, context,  D, D_index):
                        tmp_delta = {global_literal_index: [tmp_entity]}
                        ground_body(global_literal_index + 1, visited, {**delta, **tmp_delta}, {**context, **tmp_context})

                elif right_predicate in ["Bottom", "Top"]:
                    for tmp_entity, tmp_context in ground_generator(current_literal.left_literal, context, D, D_index):
                        tmp_delta = {global_literal_index: [tmp_entity]}
                        ground_body(global_literal_index + 1,visited, {**delta, **tmp_delta}, {**context, **tmp_context})

                else:
                    for left_entity, tmp_context1 in ground_generator(current_literal.left_literal, context, D):
                        for right_entity, tmp_context2 in ground_generator(current_literal.right_literal,{**context, **tmp_context1}, D):
                            tmp_delta = {global_literal_index: [left_entity, right_entity]}
                            ground_body(global_literal_index + 1, visited, {**delta, **tmp_delta}, {**context, **tmp_context1, **tmp_context2})

    if delta_old is None:
        return

    if k == 1:
        ground_body(0, -1,  {}, dict())

    else:
        for predicate in delta_old:
            for i in range(len(rule.body)):
                if not isinstance(literals[i], BinaryLiteral):
                    if literals[i].get_predicate() == predicate:
                        break
                else:
                    left_predicate = literals[i].left_literal.get_predicate()
                    right_predicate = literals[i].right_literal.get_predicate()
                    if left_predicate == predicate:
                        break
                    elif right_predicate == predicate:
                        break
            else:
                continue

            for entity in delta_old[predicate]:
                for i in range(len(rule.body)):
                    if isinstance(literals[i], BinaryLiteral):
                        left_predicate = literals[i].left_literal.get_predicate()
                        right_predicate = literals[i].right_literal.get_predicate()

                        if left_predicate == predicate:
                            context = dict()
                            for term1, term2 in zip(literals[i].left_literal.get_entity(), entity):
                                if term1.type == "constant" and term1.name != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in context and context[
                                    term1.name] != term2.name:
                                    break
                                else:
                                    if term1.type == "variable":
                                        context[term1.name] = term2.name

                            if right_predicate in ["Bottom", "Top"]:
                                # judge whether it has been visited
                                t_item = str(i) + ":"
                                for t_entity in [entity]:
                                    t_item += "-".join([str(item) for item in t_entity])
                                if t_item in history_visited:
                                    break

                                ground_body(0, i, {i: [entity]}, context)

                            else:
                                for tmp_entity, tmp_context in ground_generator(literals[i].right_literal,
                                        context, D, D_index):
                                    # judge whether it has been visited
                                    t_item = str(i) + ":"
                                    for t_entity in [entity, tmp_entity]:
                                        t_item += "-".join([str(item) for item in t_entity])
                                    if t_item in history_visited:
                                        break

                                    ground_body(0, i, {i: [entity, tmp_entity]}, {**context, **tmp_context})

                        elif right_predicate == predicate:
                            context = dict()
                            for term1, term2 in zip(literals[i].right_literal.get_entity(), entity):
                                if term1.type == "constant" and term1.name != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in context and context[
                                    term1.name] != term2.name:
                                    break
                                else:
                                    if term1.type == "variable":
                                        context[term1.name] = term2.name

                            if left_predicate in ["Bottom", "Top"]:
                                # judge whether it has been visited
                                t_item = str(i) + ":"
                                for t_entity in [entity]:
                                    t_item += "-".join([str(item) for item in t_entity])
                                if t_item in history_visited:
                                    break

                                ground_body(0, i, {i: [entity]}, context)

                            else:
                                for tmp_entity, tmp_context in ground_generator(literals[i].left_literal, context, D, D_index):
                                    # judge whether it has been visited
                                    t_item = str(i) + ":"
                                    for t_entity in [tmp_entity, entity]:
                                        t_item += "-".join([str(item) for item in t_entity])
                                    if t_item in history_visited:
                                        break

                                    ground_body(0, i, {i: [tmp_entity, entity]}, {**context, **tmp_context})

                    else:
                        if literals[i].get_predicate() == predicate:
                            context = dict()
                            for term1, term2 in zip(literals[i].get_entity(), entity):
                                if term1.type == "constant" and term1.name != term2.name:
                                    break
                                elif term1.type == "variable" and term1.name in context and context[term1.name] != term2.name:
                                    break
                                else:
                                    if term1.type == "variable":
                                        context[term1.name] = term2.name

                            t_item = str(i) + ":"
                            t_item += "-".join([str(item) for item in entity])
                            if t_item in history_visited:
                                break

                            ground_body(0, i,  {i: [entity]}, context)


if __name__ == "__main__":
    D = defaultdict(lambda: defaultdict(list))
    D["A"][tuple([Term("mike"), Term("nick")])] = [Interval(3, 4, False, False), Interval(6, 10, True, True)]
    D["B"][tuple([Term("nan")])] = [Interval(2, 8, False, False)]
    D["P"][tuple([Term("a")])] = [Interval(1, 1, False, False)]
    D_index = build_index(D)

    Delta = defaultdict(lambda: defaultdict(list))
    head = Atom("C", tuple([Term("nan")]))
    literal_a = Literal(Atom("A", tuple([Term("Y", "variable"), Term("X", "variable")])),
                        [Operator("Boxminus", Interval(1, 2, False, False))])
    literal_b = Literal(Atom("B", tuple([Term("nan")])), [Operator("Diamondminus", Interval(0, 1, False, False))])

    body = [literal_a, literal_b]

    new_literal = BinaryLiteral(Atom("A", tuple([Term("X", "variable"), Term("Y", "variable")])),
                                Atom("B", tuple([Term("nan")])), Operator("Since", Interval(1, 2, False, False)))
    body.append(new_literal)

    literal_p = Literal(Atom("P", tuple([Term("X", "variable")])), [Operator("Diamondminus", Interval(1, 1, False, False))])
    head_p = Atom("P", tuple([Term("X", "variable")]))

    rule = Rule(head_p, [literal_p])
    i = 0
    delta_old = D
    while i < 100:
        i += 1
        delta_new = defaultdict(lambda : defaultdict(list))
        seminaive_join(rule, D=D, delta_old=delta_old, delta_new=delta_new, D_index=D_index)
        print("new:")
        print_dataset(delta_new)
        for predicate in delta_new:
            if predicate not in D:
                D[predicate] = delta_new[predicate]
            else:
                for entity in delta_new[predicate]:
                    D[predicate][entity] = D[predicate][entity] + delta_new[predicate][entity]
        coalescing_d(D)
        delta_old = delta_new



















from meteor_reasoner.materialization.materialize import *
from meteor_reasoner.utils.loader import load_dataset, load_program
#
# dataset = """
# a1:name(http://www.department4.university4.edu/undergraduatestudent406,undergraduatestudent406)@(154.5,157.5]
# a1:doctoralDegreeFrom(http://www.department4.university4.edu/associateprofessor0,http://www.university181.edu)@[97.5,100.5)
# a1:publicationAuthor(http://www.department5.university4.edu/fullprofessor6/publication10,http://www.department5.university4.edu/fullprofessor6)@(169.0,171.5]
# a1:takesCourse(http://www.department0.university0.edu/undergraduatestudent433,http://www.department0.university0.edu/course2)@(34.0,35.0)
# a1:takesCourse(http://www.department0.university0.edu/undergraduatestudent5,http://www.department0.university0.edu/course21)@(124.5,126.0)
# a1:publicationAuthor(http://www.department4.university0.edu/fullprofessor1/publication4,http://www.department4.university0.edu/fullprofessor1)@[25.0,34.0]
# a1:undergraduateDegreeFrom(http://www.department9.university0.edu/lecturer1,http://www.university169.edu)@[176.5,183.5)
# a1:takesCourse(http://www.department10.university0.edu/undergraduatestudent15,http://www.department10.university0.edu/course25)@(153.5,163.5)
# a1:takesCourse(http://www.department3.university1.edu/undergraduatestudent464,http://www.department3.university1.edu/course28)@[32.0,37.0)
# a1:takesCourse(http://www.department5.university1.edu/graduatestudent75,http://www.department5.university1.edu/graduatecourse17)@[159.0,161.0]
# a1:takesCourse(http://www.department7.university1.edu/undergraduatestudent185,http://www.department7.university1.edu/course21)@[19.0,23.0)
# a1:UndergraduateStudent(http://www.department8.university1.edu/undergraduatestudent119)@[137.0,138.0)
# a1:UndergraduateStudent(http://www.department9.university1.edu/undergraduatestudent185)@(79.0,84.0]
# a1:takesCourse(http://www.department10.university1.edu/undergraduatestudent164,http://www.department10.university1.edu/course23)@[143.5,147.5)
# a1:publicationAuthor(http://www.department14.university1.edu/fullprofessor4/publication0,http://www.department14.university1.edu/fullprofessor4)@[162.5,163.5]
# a1:takesCourse(http://www.department18.university1.edu/undergraduatestudent255,http://www.department18.university1.edu/course53)@(100.5,109.5)
# a1:publicationAuthor(http://www.department18.university1.edu/fullprofessor3/publication1,http://www.department18.university1.edu/fullprofessor3)@[33.0,37.0]
# a1:memberOf(http://www.department4.university2.edu/undergraduatestudent259,http://www.department4.university2.edu)@[73.5,81.5)
# a1:FullProfessor(http://www.department5.university2.edu/fullprofessor7)@(117.5,126.5]
# a1:name(http://www.department6.university2.edu/associateprofessor8/publication1,publication1)@(74.5,77.5]
# """

D= ["R1(c1,c2)@[0,1]", "R2(c1,c2)@[1,2]", "R3(c2,c3)@[2,3]", "R5(c2)@[0,1]"]
Program = [
    "R1(X,Y):- Diamondminus[1,1]R1(X,Y)",
    "Boxplus[1,1]R5(Y):- R2(X,Y),Boxplus[1,2]R3(Y,Z)",
    "R4(X):- Diamondminus[0,1]R5(X)",
    "R6(Y):-R5(Y), Boxminus[0,2]R4(Y),R1(X,Y)"
]

D = load_dataset(D)
D_index = build_index(D)
rules = load_program(Program)
for i in range(1,4):
    print("Iteration:", i)
    print("=====================")
    print("Before D:")
    print_dataset(D)
    print("Derived facts:")
    delta_new = naive_immediate_consequence_operator(rules, D, D_index)
    print_dataset(delta_new)
    print("After D:")
    materialize(D, rules, D_index=D_index, delta_old=D, K=1)
    print_dataset(D)

D= ["R1(c1,c2)@[0,1]", "R2(c1,c2)@[1,2]", "R3(c2,c3)@[2,3]", "R5(c2)@[0,1]"]
Program = [
    "R1(X,Y):- Diamondminus[1,1]R1(X,Y)",
    "Boxplus[1,1]R5(Y):- R2(X,Y),Boxplus[1,2]R3(Y,Z)",
    "R4(X):- Diamondminus[0,1]R5(X)",
    "R6(Y):-R5(Y), Boxminus[0,2]R4(Y),R1(X,Y)"
]

D = load_dataset(D)
D_index = build_index(D)
rules = load_program(Program)
for i in range(1,4):
    print("Iteration:", i)
    print("=====================")
    print("Before D:")
    print_dataset(D)
    print("Derived facts:")
    delta_new = seminaive_immediate_consequence_operator(rules, D, D_index)
    print_dataset(delta_new)
    print("After D:")
    materialize(D, rules, D_index=D_index, delta_old=D, seminaive=True, K=1)
    print_dataset(D)






from meteor_reasoner.demo.dataset import CASES

names = ["case_1", "case_2", "case_3", "case_4", "case_5", "case_6", "case_7", "case_8"]
for name in names:
    dataset, program, _= CASES[name]

    with open(name+"_dataset.txt", "w") as file:
        for item in dataset:
            file.write(item+"\n")

    with open(name + "_program.txt", "w") as file:
        for item in program:
            file.write(item+"\n")




from meteor_reasoner.classes.literal import *
from meteor_reasoner.classes.atom import *
from meteor_reasoner.classes.interval import *
from meteor_reasoner.classes.rule import Rule
import copy


def strengthening_transformation(literal):
    new_operators = []
    if isinstance(literal, BinaryLiteral):
        operator = literal.operator
        if operator.name == "Since":
            operator.name = "Diamondminus"
        else:
            operator.name = "Diamondplus"

        if isinstance(literal.right_atom, Atom):
             literal = Literal(literal.right_atom, [operator])
        else:
            literal = literal.right_atom
            literal.operators.insert(0, operator)

    if isinstance(literal, Atom):
        return literal

    for operator in literal.operators:
        if operator.name == "Boxplus":
            operator.name = "Diamondplus"
            new_operators.append(operator)
        elif operator.name == "Boxminus":
            operator.name = "Diamondminus"
            new_operators.append(operator)
        else:
            new_operators.append(operator)

    literal.operators = new_operators[:]
    return literal


def transformation(rules):
    new_rules = []

    for rule in rules:
        head = rule.head
        new_body = []
        for literal in rule.body:
            literal = strengthening_transformation(literal)
            new_body.append(literal)
        rule = Rule(head, new_body)
        new_rules.append(rule)
    return new_rules


def get_w(rules):
    new_rules = transformation(rules[:])
    max_w = 0
    for rule in new_rules:
        head = rule.head
        w_interval = Interval(0, 0, False, False)
        if not isinstance(head, Atom):
            for operator in head.operators:
                if operator.name == "Boxminus":
                    w_interval = Interval.circle_sub(operator.interval, w_interval)
                else:
                    w_interval = Interval.add(operator.interval, w_interval)

        for literal in rule.body:
            tmp_w_interval = copy.deepcopy(w_interval)
            if not isinstance(literal, Atom):
                for operator in literal.operators:
                    if operator.name == "Diamondminus":
                        tmp_w_interval = Interval.add(operator.interval, tmp_w_interval)
                    else:
                        tmp_w_interval = Interval.circle_sub(operator.interval, tmp_w_interval)
            max_w = max(abs(tmp_w_interval.right_value), max_w)

    return max_w






import decimal


class Interval:
    def __init__(self, left_value, right_value, left_open, right_open):
        """
        An interval consists of four parts.

        Args:
            left_value (Decimal):
            right_value (Decimal):
            left_open (boolean):  True denotes open and vice versa.
            right_open (boolean): True denotes open and vice versa.
        """
        self.left_value = left_value
        self.right_value = right_value
        self.left_open = left_open
        self.right_open = right_open

    @staticmethod
    def list_union(intervals1, intervals2):
        intervals1.sort(key=lambda k: k.left_value)
        intervals2.sort(key=lambda k: k.left_value)
        new_interval_list = []
        if intervals1[0].left_value < intervals2[0].left_value:
            for interval1 in intervals1:
                for interval2 in intervals2:
                    if interval2.left_value > interval1.right_value:
                        break
                    tmp_interval = Interval.intersection(interval1, interval2)
                    if tmp_interval is not None:
                        new_interval_list.append(tmp_interval)
        else:
            for interval1 in intervals2:
                for interval2 in intervals1:
                    if interval2.left_value > interval1.right_value:
                        break
                    tmp_interval = Interval.intersection(interval1, interval2)
                    if tmp_interval is not None:
                        new_interval_list.append(tmp_interval)
        return new_interval_list

    @staticmethod
    def list_inclusion(intervals1, intervals2):
        """
        Judege whether the list of intervals(intervals1) is included by the list of intervals (intervals2)
        Args:
            intervals1: the list of intervals
            intervals2: the list of intervals

        Returns: boolean

        """
        for interval1 in intervals1:
            for interval2 in intervals2:
               if not Interval.inclusion(interval1, interval2):
                    continue
               else:
                   break
            else:
                return False

        return True

    @staticmethod
    def compare(intervals1, intervals2):
        """
        Compare two interval lists to see whether they are equal
        Args:
            intervals1 (list of Interval instances):
            intervals2 (list of Intervals instances):

        Returns:
            True or False
        """
        intervals1.sort(key=lambda item: item.left_value)
        intervals2.sort(key=lambda item: item.left_value)
        if len(intervals1) != len(intervals2):
            return False
        else:
            for v1, v2 in zip(intervals1, intervals2):
                if v1 != v2:
                    return False
            else:
                return True

    @staticmethod
    def diff(interval_src, interval_list):
        res = []
        for interval in interval_list:
            if Interval.inclusion(interval_src, interval):
                # [3,4], [[3,5],[7,10]]
                return res
            elif Interval.inclusion(interval, interval_src):
                if interval_src == interval:
                    return res

                # [2,10], [[3,5],[6,8],[9.5,10]
                if interval_src.left_value < interval.left_value:
                    res.append(Interval(interval_src.left_value, interval.left_value,
                                        interval_src.left_open, not interval.left_open))
                elif interval_src.left_value == interval.left_value and not interval_src.left_open \
                        and interval.left_open:
                    res.append(Interval(interval_src.left_value, interval.left_value,
                                        False, False))
                if interval_src.right_value == interval.right_value:
                    if interval_src.right_open == interval.right_open:
                         return res
                    elif not interval_src.right_open and interval.right_open:
                         interval_src = Interval(interval_src.right_value, interval_src.right_value,
                                                 interval_src.right_open, interval.right_open)
                    else:
                         return res
                else:
                    interval_src = Interval(interval.right_value, interval_src.right_value,
                                                not interval.right_open, interval_src.right_open)

            elif not Interval.intersection(interval, interval_src) and \
                    interval_src.right_value <= interval.left_value:
                    # [1,2], [[3,5], [7,10]
                    res.append(interval_src)
                    return res

            elif Interval.intersection(interval, interval_src):
                 # [3,5], [[4, 8]]
                 intersection_part = Interval.intersection(interval_src, interval)

                 if interval_src.right_value >= interval.right_value:
                     interval_src = Interval(intersection_part.right_value, interval_src.right_value,
                                             not intersection_part.right_open, interval_src.right_open)

                 else:
                     res.append(Interval(interval_src.left_value, intersection_part.left_value,
                                         interval_src.left_open, not intersection_part.left_open))
                     return res

        res.append(interval_src)
        return res


    @staticmethod
    def diff_list(t1_list, t2_list):
        """
        Return the different part of two interval lists
        Args:
            t1_list: a list of Intervals
            t2_list: a list of Intervals

        Returns: a list of Intervals

        """
        if len(t2_list) == 0:
            return t1_list

        res = []
        for interval1 in t1_list:
            for interval2 in t2_list:
                mover = interval1
                if Interval.inclusion(interval1, interval2):
                    mover = None
                    break
                elif Interval.intersection(interval1, interval2) is None:
                    continue
                else:
                    intersect = Interval.intersection(interval1, interval2)
                    if interval1.left_value >= intersect.left_value:
                        mover = Interval(intersect.right_value, interval1.right_value, not intersect.right_open, interval1.right_open)
                    else:
                        left = Interval(interval1.left_value, intersect.left_value, interval1.left_value, not intersect.left_open)
                        res.append(left)
                        mover = Interval(intersect.right_value, interval1.right_value, not intersect.right_open, interval1.right_open)
            if mover is not None:
                  res.append(mover)
        return res


    @staticmethod
    def union(v1, v2):
        """
        Return the union of v1 and v2 if v1 and v2 have an intersection part, else return None.
        Args:
            v1 (an Interval instance):
            v2 (an Interval instance):

        Returns:
            None or a newly-created interval instance
        """
        if v1.left_value > v2.left_value:
            v1, v2 = v2, v1

        if v1.left_value == v2.left_value:
            left_value = v1.left_value
            if v1.left_open != v2.left_open:
                left_open = False
            else:
                left_open = v1.left_open

            right_value = max(v1.right_value, v2.right_value)
            if v1.right_value == v2.right_value and v1.right_open != v2.right_open:
                right_open = False
            else:
                if v1.right_value == right_value:
                    right_open = v1.right_open
                else:
                    right_open = v2.right_open

            return Interval(left_value, right_value, left_open, right_open)

        elif v2.left_value == v1.right_value and v2.left_open and v1.right_open:
            return None

        else:
            if v2.left_value <= v1.right_value:
                left_value = v1.left_value
                left_open = v1.left_open
                right_value = max(v1.right_value, v2.right_value)
                if v1.right_value == v2.right_value and v1.right_open != v2.right_open:
                    right_open = False
                else:
                    if v1.right_value == right_value:
                        right_open = v1.right_open
                    else:
                        right_open = v2.right_open

                return Interval(left_value, right_value, left_open, right_open)
            else:
                return None

    @staticmethod
    def inclusion(v1, v2):
        if v1.left_value < v2.left_value or v1.right_value > v2.right_value:
            return False
        else:
            if v1.left_value == v2.left_value and v2.left_open and not v1.left_open:
                return False
            elif v1.right_value == v2.right_value and v2.right_open and not v1.right_open:
                return False
            else:
                return True

    @staticmethod
    def intersection(v1, v2):
        """
        Return the intersection of v1 and v2
        Args:
            v1 (an Interval instance):
            v2 (an Interval instance):

        Returns:
            None or a newly-created interval instance
        """
        if v1.left_value > v2.left_value:
            left_value = v1.left_value
            left_open = v1.left_open

        elif v1.left_value < v2.left_value:
            left_value = v2.left_value
            left_open = v2.left_open
        else:
            left_value = v2.left_value
            if v1.left_open != v2.left_open:
                left_open = True
            else:
                left_open = v2.left_open

        if v1.right_value < v2.right_value:
            right_value = v1.right_value
            right_open = v1.right_open
        elif v1.right_value > v2.right_value:
            right_value = v2.right_value
            right_open = v2.right_open
        else:
            right_value = v2.right_value
            if v1.right_open != v2.right_open:
                right_open =True
            else:
                right_open = v2.right_open

        if not Interval.is_valid_interval(left_value, right_value, left_open, right_open):
            return None
        else:
            return Interval(left_value, right_value, left_open, right_open)

    @staticmethod
    def is_valid_interval(left_value, right_value, left_open, right_open):
        """
        Check whether the given four values can form a valid Interval.
        Args:
            left_value (float):
            right_value (float):
            left_open (boolean):
            right_open (boolean):

        Returns:
            boolean
        """
        if left_value == right_value and left_open and right_open:
            return False
        elif left_value > right_value:
            return False
        elif left_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")] and not left_open:
            return False
        elif right_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")] and not right_open:
            return False
        elif left_value == right_value and left_open != right_open:
            return False

        return True

    def __str__(self):
        str_output = ""
        if self.left_open:
            str_output += "("
        else:
            str_output += "["
        if self.left_value in [decimal.Decimal("-inf")]:
            str_output += "-inf"
        elif self.left_value in [decimal.Decimal("inf")]:
            str_output += "+inf"
        else:
            str_output += str(self.left_value)

        str_output += ","

        if self.right_value in [decimal.Decimal("-inf")]:
            str_output += "-inf"
        elif self.right_value in [decimal.Decimal("inf")]:
            str_output += "+inf"
        else:
            str_output += str(self.right_value)

        if self.right_open:
            str_output += ")"
        else:
            str_output += "]"

        return str_output

    def __hash__(self):
        return hash(str(self.left_open) + str(self.left_value) + ":" + str(self.right_value) + str(self.right_open))

    def __eq__(self, other):
        if isinstance(other, Interval):
            return self.left_value == other.left_value and self.right_value == other.right_value \
                   and self.left_open == other.left_open and self.right_open == other.right_open

    @staticmethod
    def sub(v1, v2):
        """
        Implement v1- sub v2+, and v1+ sub v2-, where v- represents the left value of the Interval instance and v+
        denotes the right value of the Interval instance.
        Args:
            v1 (an Interval instance):
            v2 (an Interval instance):

        Returns:
            a newly-created Interval instance
        """
        if v1.left_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")]:
            left_value = v1.left_value
        else:
            left_value = v1.left_value - v2.right_value
        if v1.right_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")]:
            right_value = v1.right_value
        else:
            right_value = v1.right_value - v2.left_value

        left_open, right_open = False, False

        if left_value == decimal.Decimal("-inf") or (v1.left_open or v2.right_open):
            left_open = True

        if right_value == decimal.Decimal("inf") or (v1.right_open or v2.left_open):
            right_open = True

        return Interval(left_value, right_value, left_open, right_open)

    @staticmethod
    def circle_sub(v1, v2):
        """
        Implement v1- sub v2-, and v1+ sub v2+, where v- represents the left value of the Interval instance and v+
        denotes the right value of the Interval instance.
        Args:
           v1 (an Interval instance):
           v2 (an Interval instance):

        Returns:
           a newly-created Interval instance
        """
        if v1.left_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")]:
            left_value = v1.left_value
        else:
            left_value = v1.left_value - v2.left_value
        if v1.right_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")]:
            right_value = v1.right_value
        else:
            right_value = v1.right_value - v2.right_value

        left_open, right_open = False, False

        if left_value == decimal.Decimal("-inf") or (v1.left_open and not v2.left_open):
            left_open = True

        if right_value == decimal.Decimal("inf") or (v1.right_open and not v2.right_open):
            right_open = True

        return Interval(left_value, right_value, left_open, right_open)

    @staticmethod
    def add(v1, v2):
        """
        Implement v1- add v2-, and v1+ add v2+, where v- represents the left value of the Interval instance and v+
        denotes the right value of the Interval instance.
        Args:
           v1 (an Interval instance):
           v2 (an Interval instance):

        Returns:
           a newly-created Interval instance
        """
        if v1.left_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")]:
            left_value = v1.left_value
        else:
            left_value = v1.left_value + v2.left_value
        if v1.right_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")]:
            right_value = v1.right_value
        else:
            right_value = v1.right_value + v2.right_value

        left_open, right_open = False, False

        if left_value == decimal.Decimal("-inf") or (v1.left_open or v2.left_open):
            left_open = True

        if right_value == decimal.Decimal("inf") or (v1.right_open or v2.right_open):
            right_open = True

        return Interval(left_value, right_value, left_open, right_open)

    @staticmethod
    def circle_add(v1, v2):
        """
        Implement v1- add v2+, and v1+ add v2-, where v- represents the left value of the Interval instance and v+
        denotes the right value of the Interval instance.
        Args:
           v1 (an Interval instance):
           v2 (an Interval instance):

        Returns:
           a newly-created Interval instance
        """
        if v1.left_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")]:
            left_value = v1.left_value
        else:
            left_value = v1.left_value + v2.right_value
        if v1.right_value in [decimal.Decimal("-inf"), decimal.Decimal("inf")]:
            right_value = v1.right_value
        else:
            right_value = v1.right_value + v2.left_value
        left_open, right_open = False, False

        if left_value == decimal.Decimal("-inf") or (v1.left_open and not v2.right_open):
            left_open = True

        if right_value == decimal.Decimal("inf") or (v1.right_open and not v2.left_open):
            right_open = True

        return Interval(left_value, right_value, left_open, right_open)


if __name__ == "__main__":

    # a1 = Interval(1.5, 2.5, True, False)
    # a2 = Interval(2.5, 4, True, False)
    # print(Interval.intersection(a1, a2))
    # exit()

    a = [Interval(0, 0, False, False), Interval(2.0, 5, False, False)]
    b = [Interval(0.0, 0.0, False, False)]

    print(Interval.list_inclusion(b, a))
    exit()

    c = Interval.diff_list(b, a)
    # ['[2.5,4.0]']
    c = [Interval(0, 12,   False, False), Interval(21, 64, False, False)]

    print([str(item) for item in Interval.diff(b, a)])
    print([str(item) for item in Interval.list_union(c, a)])
    print(Interval.list_inclusion(a, c))
